# -*- coding: utf-8 -*-
from __future__ import division, print_function, absolute_import


import unittest
import numpy as np
from anhima.gt import as_012, as_n_alt, as_allele_counts, pack_diploid, \
    unpack_diploid, BHOM00, BHET01, BHOM11, BMISSING


class TestTransformations(unittest.TestCase):

    def setUp(self):
        self.genotypes = np.array(
            [[[0, 0], [0, 1]],
             [[1, 1], [-1, -1]],
             [[0, 2], [1, 2]],
             [[2, 2], [2, 0]]],
            dtype='i1'
        )

    def test_as_012(self):
        expect = np.array(
            [[0, 1],
             [2, -1],
             [1, 1],
             [2, 1]]
        )
        actual = as_012(self.genotypes)
        assert np.array_equal(expect, actual)

    def test_as_nalt(self):
        expect = np.array(
            [[0, 1],
             [2, 0],
             [1, 2],
             [2, 1]]
        )
        actual = as_n_alt(self.genotypes)
        assert np.array_equal(expect, actual)

    def test_as_allele_counts(self):
        expect = np.array(
            [[[2, 0, 0], [1, 1, 0]],
             [[0, 2, 0], [0, 0, 0]],
             [[1, 0, 1], [0, 1, 1]],
             [[0, 0, 2], [1, 0, 1]]]
        )
        actual = as_allele_counts(self.genotypes)
        assert np.array_equal(expect, actual)


class TestPackUnpackDiploid(unittest.TestCase):

    def test_pack_unpack_diploid(self):
        """Check that genotypes survive pack/unpack round trip."""

        genotypes = np.array([[[0, 0], [0, 1]],
                              [[1, 1], [-1, -1]],
                              [[6, 6], [7, 7]],
                              [[10, 10], [14, 14]]], dtype='i1')
        packed = pack_diploid(genotypes)
        self.assertEqual(BHOM00, packed[0, 0])
        self.assertEqual(BHET01, packed[0, 1])
        self.assertEqual(BHOM11, packed[1, 0])
        self.assertEqual(BMISSING, packed[1, 1])
        actual = unpack_diploid(packed)
        self.assertTrue(np.array_equal(genotypes, actual))

    def test_pack_diploid_allele_range(self):
        """Check that alleles outside of the supported range raise an error."""

        genotypes = np.array([[[-2, -2]]], dtype='i1')
        self.assertRaises(AssertionError,
                          pack_diploid,
                          genotypes)

        genotypes = np.array([[[15, 15]]], dtype='i1')
        self.assertRaises(AssertionError,
                          pack_diploid,
                          genotypes)

    def test_pack_diploid_ref_zero(self):
        """Check that hom ref calls are encoded as 0 for sparse matrices."""

        genotypes = np.array([[[0, 0]]], dtype='i1')
        packed = pack_diploid(genotypes)
        self.assertEqual(0, packed[0, 0])



# -*- coding: utf-8 -*-
"""
Miscellaneous utilities.

"""


from __future__ import division, print_function, absolute_import
from anhima.compat import range


# third party dependencies
import numpy as np
import pandas


def block_take2d(dataset, row_indices, col_indices=None, block_size=None):
    """Select rows and optionally columns from a Numpy array or HDF5
    dataset with 2 or more dimensions.

    Parameters
    ----------

    dataset : array_like or HDF5 dataset
        The input dataset.
    row_indices : sequence of ints
        The indices of the selected rows. N.B., will be sorted in ascending
        order.
    col_indices : sequence of ints, optional
        The indices of the selected columns. If not provided, all columns
        will be returned.
    block_size : int, optional
        The size (in number of rows) of the block of data to process at a time.

    Returns
    -------

    out : ndarray
        An array containing the selected rows and columns.

    See Also
    --------

    anhima.util.block_compress2d, anhima.h5.take2d_pointsel

    Notes
    -----

    This function is mainly a work-around for the fact that fancy indexing via
    h5py is currently slow, and fancy indexing along more than one axis is not
    supported. The function works by reading the entire dataset in blocks of
    `block_size` rows, and processing each block in memory using numpy.

    """

    # N.B., make sure row_indices are sorted
    row_indices = np.asarray(row_indices)
    row_indices.sort()

    # how many rows are we selecting?
    n_rows_in = dataset.shape[0]
    n_rows_out = len(row_indices)

    # how many columns are we selecting?
    n_cols_in = dataset.shape[1]
    if col_indices:
        n_cols_out = len(col_indices)
    else:
        n_cols_out = n_cols_in

    # setup output array
    out_shape = (n_rows_out, n_cols_out) + dataset.shape[2:]
    out = np.empty(out_shape, dtype=dataset.dtype)

    # determine block size
    if block_size is None:
        if hasattr(dataset, 'chunks') and dataset.chunks is not None:
            # use dataset chunk height
            block_size = dataset.chunks[0]
        else:
            # use arbitrary number
            block_size = 1000

    # iterate block-wise
    offset = 0
    for block_start in range(0, n_rows_in, block_size):
        block_stop = min(block_start+block_size, n_rows_in)

        # how many indices to process in this block?
        i = np.searchsorted(row_indices, block_start)
        j = np.searchsorted(row_indices, block_stop)
        n = j-i
        ridx = row_indices[i:j]

        # only do anything if there are indices for this block
        if n:

            # load data for this block
            a = dataset[block_start:block_stop]

            # take rows
            b = np.take(a, ridx-block_start, axis=0)

            # take columns
            if col_indices:
                b = np.take(b, col_indices, axis=1)

            # store output
            out[offset:offset+n, ...] = b

            # keep track of offset
            offset += n

    return out


def block_compress2d(dataset, row_condition, col_condition=None,
                     block_size=None):
    """Select rows and optionally columns from a Numpy array or HDF5
    dataset with 2 or more dimensions.

    Parameters
    ----------

    dataset : array_like or HDF5 dataset
        The input dataset.
    row_condition : array_like, bool
        A boolean array indicating the selected rows.
    col_indices : array_like, bool, optonal
        A boolean array indicated the selected columns. If not provided,
        all columns will be returned.
    block_size : int, optional
        The size (in number of rows) of the block of data to process at a time.

    Returns
    -------

    out : ndarray
        An array containing the selected rows and columns.

    See Also
    --------

    anhima.util.block_take2d, anhima.h5.take2d_pointsel

    Notes
    -----

    This function is mainly a work-around for the fact that fancy indexing via
    h5py is currently slow, and fancy indexing along more than one axis is not
    supported. The function works by reading the entire dataset in blocks of
    `block_size` rows, and processing each block in memory using numpy.

    """

    row_indices = np.nonzero(row_condition)[0]
    col_indices = np.nonzero(col_condition)[0] if col_condition is not None \
        else None
    return block_take2d(dataset, row_indices, col_indices,
                        block_size=block_size)


def block_apply(f, dataset, block_size=None, out=None):
    """Apply function `f` to `dataset` split along the first axis into
    contiguous slices of `block_size`. The result should be equivalent to
    calling ``f(dataset)`` directly, however may require less total memory,
    especially if `dataset` is an HDF5 dataset.

    Parameters
    ----------

    f : function
        The function to apply.
    dataset : array_like or HDF5 dataset
        The input dataset.
    block_size : int, optional
        The size (in number of items along `axis`) of the blocks passed to `f`.
    out : array_like or HDF5 dataset, optional
        If given, used to store the output.

    Returns
    -------

    out : ndarray
        The result of applying `f` to `dataset` blockwise.

    """

    # determine block size
    if block_size is None:
        if hasattr(dataset, 'chunks') and dataset.chunks is not None:
            # use dataset chunk size along slice axis
            block_size = dataset.chunks[0]
        else:
            # use arbitrary number
            block_size = 1000

    # determine total size along slice axis
    dim_size = dataset.shape[0]

    # iterate over blocks
    for block_start in range(0, dim_size, block_size):
        block_stop = min(block_start + block_size, dim_size)

        # load input block
        x = dataset[block_start:block_stop, ...]

        # compute output block
        y = f(x)

        if out is None:
            # initialise output array
            out_shape = list(y.shape)
            out_shape[0] = dim_size
            out = np.empty(out_shape, y.dtype)

        # store output block
        out[block_start:block_stop, ...] = y

    return out


def state_transitions(x, states):
    """Find state transitions.

    Parameters
    ----------

    x : array_like
        One-dimensional array of state values.
    states : set-like
        Set of states to consider (other state values are ignored).

    Returns
    -------

    switch_points : ndarray
        Two-dimensional array of switch points, where each row stores a pair
        of indices corresponding to the indices either side of a switch.
    transitions : ndarray
        Two-dimensional array of transitions, where each row stores a pair
        of state values corresponding to states either side of a switch.

    """

    # check inputs
    x = np.asarray(x)
    assert x.ndim == 1

    # setup output variables
    switch_points = list()
    transitions = list()

    # utility variables
    prv = None
    prv_idx = None

    # iterate over state values
    for cur_idx, cur in enumerate(x):

        if cur not in states:
            # ignore
            pass

        else:

            if prv is None:
                # first informative state
                pass

            elif cur != prv:
                # record a state transition
                switch = prv_idx, cur_idx
                switch_points.append(switch)
                transition = prv, cur
                transitions.append(transition)

            # advance
            prv = cur
            prv_idx = cur_idx

    return np.array(switch_points), np.array(transitions)


def tabulate_state_transitions(x, states, pos):
    """TODO

    """

    # check inputs
    x = np.asarray(x)
    assert x.ndim == 1
    pos = np.asarray(pos)
    assert pos.ndim == 1
    assert x.size == pos.size

    switch_points, transitions = state_transitions(x, states)
    switch_positions = np.take(pos, switch_points)
    data = [('lidx', switch_points[:, 0]),
            ('ridx', switch_points[:, 1]),
            ('lpos', switch_positions[:, 0]),
            ('rpos', switch_positions[:, 1]),
            ('lval', transitions[:, 0]),
            ('rval', transitions[:, 1])]
    df = pandas.DataFrame.from_items(data)
    return df


def tabulate_state_blocks(x, states, pos):
    """TODO

    """

    # check inputs
    x = np.asarray(x)
    assert x.ndim == 1
    n = x.size
    pos = np.asarray(pos)
    assert pos.ndim == 1
    assert pos.size == n

    blocks = list()

    df_switches = tabulate_state_transitions(x, states, pos)
    for i, switch in enumerate(df_switches.values):
        lidx, ridx, lpos, rpos, lval, rval = switch
        block_stop_min_idx = lidx
        block_stop_max_idx = ridx
        block_stop_min_pos = lpos
        block_stop_max_pos = rpos
        block_state = lval

        if i == 0:
            # special case the first switch
            block_start_min_idx = 0
            block_start_max_idx = 0
            block_start_min_pos = pos[0]
            block_start_max_pos = pos[0]
            block_is_marginal = True
            block_size_min = block_stop_min_idx
            block_size_max = block_stop_max_idx - 1
        else:
            previous_switch = df_switches.iloc[i-1]
            block_start_min_idx = previous_switch.lidx
            block_start_max_idx = previous_switch.ridx
            block_start_min_pos = previous_switch.lpos
            block_start_max_pos = previous_switch.rpos
            block_is_marginal = False
            block_size_min = block_stop_min_idx - block_start_max_idx + 1
            block_size_max = block_stop_max_idx - block_start_min_idx - 1

        y = x[block_start_max_idx:block_stop_min_idx+1]
        block_support = np.count_nonzero(y == block_state)
        block_length_min = block_stop_min_pos - block_start_max_pos
        block_length_max = block_stop_max_pos - block_start_min_pos

        block = (block_start_min_idx, block_start_max_idx,
                 block_stop_min_idx, block_stop_max_idx,
                 block_start_min_pos, block_start_max_pos,
                 block_stop_min_pos, block_stop_max_pos,
                 block_state, block_support,
                 block_size_min, block_size_max,
                 block_length_min, block_length_max, block_is_marginal)
        blocks.append(block)

    # special case the last block
    previous_switch = df_switches.iloc[-1]
    block_start_min_idx = previous_switch.lidx
    block_start_max_idx = previous_switch.ridx
    block_start_min_pos = previous_switch.lpos
    block_start_max_pos = previous_switch.rpos
    block_is_marginal = True
    block_stop_min_idx = n-1
    block_stop_max_idx = n-1
    block_stop_min_pos = pos[-1]
    block_stop_max_pos = pos[-1]
    block_state = previous_switch.rval
    y = x[block_start_max_idx:block_stop_min_idx+1]
    block_support = np.count_nonzero(y == block_state)
    block_size_min = block_stop_min_idx - block_start_max_idx + 1
    block_size_max = block_stop_max_idx - block_start_min_idx
    block_length_min = block_stop_min_pos - block_start_max_pos
    block_length_max = block_stop_max_pos - block_start_min_pos

    block = (block_start_min_idx, block_start_max_idx,
             block_stop_min_idx, block_stop_max_idx,
             block_start_min_pos, block_start_max_pos,
             block_stop_min_pos, block_stop_max_pos,
             block_state, block_support,
             block_size_min, block_size_max,
             block_length_min, block_length_max, block_is_marginal)
    blocks.append(block)

    columns = ('start_min_idx', 'start_max_idx',
               'stop_min_idx', 'stop_max_idx',
               'start_min_pos', 'start_max_pos',
               'stop_min_pos', 'stop_max_pos',
               'state', 'support',
               'size_min', 'size_max',
               'length_min', 'length_max', 'is_marginal')
    df = pandas.DataFrame.from_records(blocks, columns=columns)
    return df



# -*- coding: utf-8 -*-
"""
Allele frequency calculations.

See also the examples at:

- http://nbviewer.ipython.org/github/alimanfoo/anhima/blob/master/examples/af.ipynb

"""  # noqa


from __future__ import division, print_function, absolute_import


# third party dependencies
import numpy as np


# internal dependencies
import anhima


def _check_genotypes(genotypes):
    """
    Internal function to check the genotypes input argument meets
    expectations.

    """

    genotypes = np.asarray(genotypes)
    assert genotypes.ndim >= 2

    if genotypes.ndim == 2:
        # assume haploid, add ploidy dimension
        genotypes = genotypes[..., np.newaxis]

    return genotypes


def is_variant(genotypes):
    """Find variants with at least one non-reference allele observation.

    Parameters
    ----------

    genotypes : array_like
        An array of shape (n_variants, n_samples, ploidy) where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).

    Returns
    -------

    is_variant : ndarray, bool
        An array of shape (n_variants,) where an element is True if there
        are at least `min_ac` non-reference alleles found for the corresponding
        variant.

    Notes
    -----

    Applicable to polyploid genotype calls.

    Applicable to multiallelic variants.

    """

    # check inputs
    genotypes = _check_genotypes(genotypes)

    # aggregate over samples and ploidy dimensions
    out = np.sum(genotypes > 0, axis=(1, 2)) >= 1

    return out


def count_variant(genotypes):
    """Count variants with at least one non-reference allele observed.

    Parameters
    ----------

    genotypes : array_like
        An array of shape (n_variants, n_samples, ploidy) where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).
    min_ac : int, optional
        The minimum number of non-reference alleles required to consider
        variant.

    Returns
    -------

    n : int
        The number of variants.

    Notes
    -----

    Applicable to polyploid genotype calls.

    Applicable to multiallelic variants.

    """

    return np.count_nonzero(is_variant(genotypes))


def is_non_variant(genotypes):
    """Find variants with no non-reference alleles.

    Parameters
    ----------

    genotypes : array_like
        An array of shape (n_variants, n_samples, ploidy) where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).

    Returns
    -------

    is_non_variant : ndarray, bool
        An array of shape (n_variants,) where an element is True if there
        are no non-reference alleles found for the corresponding variant.

    Notes
    -----

    Applicable to polyploid genotype calls.

    Applicable to multiallelic variants.

    """

    # check inputs
    genotypes = _check_genotypes(genotypes)

    # aggregate over samples and ploidy dimensions
    out = np.all(genotypes <= 0, axis=(1, 2))

    return out


def count_non_variant(genotypes):
    """Count variants with no non-reference alleles.

    Parameters
    ----------

    genotypes : array_like
        An array of shape (n_variants, n_samples, ploidy) where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).

    Returns
    -------

    n : int
        The number of variants.

    Notes
    -----

    Applicable to polyploid genotype calls.

    Applicable to multiallelic variants.

    """

    return np.count_nonzero(is_non_variant(genotypes))


def is_singleton(genotypes, allele=1):
    """Find variants with only a single instance of `allele` observed.

    Parameters
    ----------

    genotypes : array_like
        An array of shape (n_variants, n_samples, ploidy) where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).
    allele : int, optional
        The allele to find singletons of.

    Returns
    -------

    is_singleton : ndarray, bool
        An array of shape (n_variants,) where an element is True if there
        is a single instance of `allele` observed.

    Notes
    -----

    Applicable to polyploid genotype calls.

    Applicable to multiallelic variants, but note this function checks for a
    specific `allele`.

    """

    # check inputs
    genotypes = _check_genotypes(genotypes)

    # aggregate over samples and ploidy dimensions
    out = np.sum(genotypes == allele, axis=(1, 2)) == 1

    return out


def count_singletons(genotypes, allele=1):
    """Count variants with only a single instance of `allele` observed.

    Parameters
    ----------

    genotypes : array_like
        An array of shape (n_variants, n_samples, ploidy) where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).
    allele : int, optional
        The allele to find singletons of.

    Returns
    -------

    n : int
        The number of variants.

    Notes
    -----

    Applicable to polyploid genotype calls.

    Applicable to multiallelic variants, but note this function checks for a
    specific `allele`.

    """

    return np.count_nonzero(is_singleton(genotypes, allele))


def is_doubleton(genotypes, allele=1):
    """Find variants with only two instances of `allele` observed.

    Parameters
    ----------

    genotypes : array_like
        An array of shape (n_variants, n_samples, ploidy) where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).
    allele : int, optional
        The allele to find doubletons of.

    Returns
    -------

    is_doubleton : ndarray, bool
        An array of shape (n_variants,) where an element is True if there
        are exactly two instances of `allele` observed.

    Notes
    -----

    Applicable to polyploid genotype calls.

    Applicable to multiallelic variants, but note this function checks for a
    specific `allele`.

    """

    # check inputs
    genotypes = _check_genotypes(genotypes)

    # aggregate over samples and ploidy dimensions
    out = np.sum(genotypes == allele, axis=(1, 2)) == 2

    return out


def count_doubletons(genotypes, allele=1):
    """Count variants with only two instances of `allele` observed.

    Parameters
    ----------

    genotypes : array_like
        An array of shape (n_variants, n_samples, ploidy) where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).
    allele : int, optional
        The allele to find doubletons of.

    Returns
    -------

    n : int
        The number of variants.

    Notes
    -----

    Applicable to polyploid genotype calls.

    Applicable to multiallelic variants, but note this function checks for a
    specific `allele`.

    """

    return np.count_nonzero(is_doubleton(genotypes, allele))


def allele_number(genotypes):
    """Count the number of non-missing allele calls per variant.

    Parameters
    ----------

    genotypes : array_like
        An array of shape (n_variants, n_samples, ploidy) where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).

    Returns
    -------

    an : ndarray, int
        An array of shape (n_variants,) counting the total number of
        non-missing alleles observed.

    Notes
    -----

    Applicable to polyploid genotype calls.

    Applicable to multiallelic variants.

    """

    # check inputs
    genotypes = _check_genotypes(genotypes)

    # aggregate over samples and ploidy dimensions
    an = np.sum(genotypes >= 0, axis=(1, 2))

    return an


def allele_count(genotypes, allele=1):
    """Calculate number of observations of the given allele per variant.

    Parameters
    ----------

    genotypes : array_like
        An array of shape (n_variants, n_samples, ploidy) where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).
    allele : int, optional
        The allele to count.

    Returns
    -------

    ac : ndarray, int
        An array of shape (n_variants,) counting the number of
        times the given `allele` was observed.

    Notes
    -----

    Applicable to polyploid genotype calls.

    Applicable to multiallelic variants, but note that this function
    calculates the frequency of a specific `allele`.

    """

    # check inputs
    genotypes = _check_genotypes(genotypes)

    # aggregate over samples and ploidy dimensions
    ac = np.sum(genotypes == allele, axis=(1, 2))

    return ac


def allele_frequency(genotypes, allele=1):
    """Calculate frequency of the given allele per variant.

    Parameters
    ----------

    genotypes : array_like
        An array of shape (n_variants, n_samples, ploidy) where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).
    allele : int, optional
        The allele to calculate the frequency of.

    Returns
    -------

    an : ndarray, int
        An array of shape (n_variants,) counting the total number of
        non-missing alleles observed.
    ac : ndarray, int
        An array of shape (n_variants,) counting the number of
        times the given `allele` was observed.
    af : ndarray, float
        An array of shape (n_variants,) containing the allele frequency.

    Notes
    -----

    Applicable to polyploid genotype calls.

    Applicable to multiallelic variants, but note that this function
    calculates the frequency of a specific `allele`.

    """

    # count non-missing alleles
    an = allele_number(genotypes)

    # count alleles
    ac = allele_count(genotypes, allele=allele)

    # calculate allele frequency, accounting for missingness
    err = np.seterr(invalid='ignore')
    af = np.where(an > 0, ac / an, 0)
    np.seterr(**err)

    return an, ac, af


def allele_counts(genotypes, alleles=None):
    """Calculate allele counts per variant.

    Parameters
    ----------

    genotypes : array_like
        An array of shape (n_variants, n_samples, ploidy) where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).
    alleles : sequence of ints, optional
        The alleles to count. If not specified, all alleles will be counted.

    Returns
    -------

    ac : ndarray, int
        An array of shape (n_variants, n_alleles) counting the number of
        times the given `alleles` were observed.

    Notes
    -----

    Applicable to polyploid genotype calls.

    Applicable to multiallelic variants.

    """

    # check input
    genotypes = _check_genotypes(genotypes)

    # determine number of variants
    n_variants = genotypes.shape[0]

    # if alleles not specified, count all alleles
    if alleles is None:
        m = np.amax(genotypes)
        alleles = range(m+1)

    # count alleles
    ac = np.zeros((n_variants, len(alleles)), dtype='i4')
    for i, allele in enumerate(alleles):
        np.sum(genotypes == allele, axis=(1, 2), out=ac[:, i])

    return ac


def allele_frequencies(genotypes, alleles=None):
    """Calculate allele frequencies per variant.

    Parameters
    ----------

    genotypes : array_like
        An array of shape (n_variants, n_samples, ploidy) where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).
    alleles : sequence of ints, optional
        The alleles to calculate the frequency of. If not specified, all
        alleles will be counted.

    Returns
    -------

    an : ndarray, int
        An array of shape (n_variants,) counting the total number of
        non-missing alleles observed.
    ac : ndarray, int
        An array of shape (n_variants, n_alleles) counting the number of
        times the given `alleles` were observed.
    af : ndarray, float
        An array of shape (n_variants, n_alleles) containing the allele
        frequencies.

    Notes
    -----

    Applicable to polyploid genotype calls.

    Applicable to multiallelic variants.

    """

    # count non-missing alleles
    an = allele_number(genotypes)[:, np.newaxis]

    # count alleles
    ac = allele_counts(genotypes, alleles=alleles)

    # calculate allele frequencies, accounting for missingness
    err = np.seterr(invalid='ignore')
    af = np.where(an > 0, ac / an, 0)
    np.seterr(**err)

    return an, ac, af


def allelism(genotypes):
    """Determine the number of distinct alleles found for each variant.

    Parameters
    ----------

    genotypes : array_like
        An array of shape (n_variants, n_samples, ploidy) where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).

    Returns
    -------

    n : ndarray, int
        An array of shape (n_variants,) where an element holds the allelism
        of the corresponding variant.

    See Also
    --------

    max_allele

    Notes
    -----

    Applicable to polyploid genotype calls.

    Applicable to multiallelic variants.

    """

    # check inputs
    genotypes = _check_genotypes(genotypes)

    # calculate allele counts
    ac = allele_counts(genotypes)

    # count alleles present
    n = np.sum(ac > 0, axis=1)

    return n


def is_non_segregating(genotypes, allele=None):
    """Find non-segregating variants (fixed for a single allele).

    Parameters
    ----------

    genotypes : array_like
        An array of shape (n_variants, n_samples, ploidy) where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).
    allele : int, optional
        If given, find variants fixed with respect to `allele`. Otherwise
        find variants fixed for any allele.

    Returns
    -------

    is_non_segregating : ndarray, bool
        An array of shape (n_variants,) where an element is True if all
        genotype calls for the corresponding variant are either missing or
        equal to the same allele.

    Notes
    -----

    Applicable to polyploid genotype calls.

    Applicable to multiallelic variants.

    """

    # check inputs
    genotypes = _check_genotypes(genotypes)

    if allele is None:

        # count distinct alleles
        n_alleles = allelism(genotypes)

        # find fixed variants
        out = n_alleles == 1

    else:

        # find fixed variants with respect to a specific allele
        out = np.all((genotypes < 0) | (genotypes == allele), axis=(1, 2))

    return out


def count_non_segregating(genotypes, allele=None):
    """Count non-segregating variants.

    Parameters
    ----------

    genotypes : array_like
        An array of shape (n_variants, n_samples, ploidy) where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).
    allele : int, optional
        If given, find variants fixed with respect to `allele`.

    Returns
    -------

    n : int
        The number of variants.

    Notes
    -----

    Applicable to polyploid genotype calls.

    Applicable to multiallelic variants.

    """

    return np.count_nonzero(is_non_segregating(genotypes, allele=allele))


def is_segregating(genotypes):
    """Find segregating variants (where more than one allele is observed).

    Parameters
    ----------

    genotypes : array_like
        An array of shape (n_variants, n_samples, ploidy) where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).

    Returns
    -------

    is_segregating : ndarray, bool
        An array of shape (n_variants,) where an element is True if more
        than one allele is found for the given variant.

    Notes
    -----

    Applicable to polyploid genotype calls.

    Applicable to multiallelic variants.

    """

    # check inputs
    genotypes = _check_genotypes(genotypes)

    # count distinct alleles
    n_alleles = allelism(genotypes)

    # find segregating variants
    out = n_alleles > 1

    return out


def count_segregating(genotypes):
    """Count segregating variants.

    Parameters
    ----------

    genotypes : array_like
        An array of shape (n_variants, n_samples, ploidy) where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).

    Returns
    -------

    n : int
        The number of variants.

    Notes
    -----

    Applicable to polyploid genotype calls.

    Applicable to multiallelic variants.

    """

    return np.count_nonzero(is_segregating(genotypes))


def maximum_likelihood_ancestry(genotypes, qa, qb, filter_size=0):
    """Given alternate allele frequencies in two populations `qa` and `qb`,
    predict the ancestry for a set of `genotypes`.

    Parameters
    ----------

    genotypes : array_like
        An array of diploid genotype calls of shape (n_variants, n_samples,
        2) where each element of the array is an integer corresponding to an
        allele index (-1 = missing, 0 = reference allele, 1 = first alternate
        allele, 2 = second alternate allele, etc.).
    qa : array_like, float
        A 1-dimensional array of shape (n_variants, ) containing alternate
        allele frequencies for population A.
    qb : array_like, float
        A 1-dimensional array of shape (n_variants, ) containing alternate
        allele frequencies for population B.
    filter_size : int, optional
        Sum likelihoods in a moving window of size `filter_size`.

    Returns
    -------

    ancestry : ndarray, int, shape (n_variants, n_samples)
        An array containing the ancestry predictions, where 0 = AA (both
        alleles derive from population A), 1 = AB (hybrid ancestry) and 2 =
        BB (both alleles derive from population B), and -1 = ambiguous (models
        are equally likely).
    confidence : ndarray, float, shape (n_variants, n_samples)
        The confidence in the ancestry prediction (natural logarithm of the
        likelihood ratio for the two most likely models).

    Notes
    -----

    Where allele frequencies are similar between populations A and B,
    ancestry predictions will have low confidence, because different ancestry
    models will have similar likelihoods. Greater confidence will be obtained by
    filtering variants to select those where the difference in allele
    frequencies is greater. E.g.::

        >>> flt = np.abs(qa - qb) > .5
        >>> genotypes_flt = genotypes[flt]
        >>> qa_flt = qa[flt]
        >>> qb_flt = qb[flt]
        >>> ancestry, confidence = maximum_likelihood_ancestry(genotypes_flt, qa_flt, qb_flt)

    """  # noqa

    # check inputs
    genotypes = _check_genotypes(genotypes)
    # require biallelic genotypes
    assert np.amax(genotypes) < 2
    n_variants, n_samples, ploidy = genotypes.shape
    # require diploid genotypes
    assert ploidy == 2
    qa = np.asarray(qa)
    qb = np.asarray(qb)
    assert qa.ndim == qb.ndim == 1
    assert n_variants == qa.shape[0] == qb.shape[0]

    # calculate reference allele frequencies, assuming biallelic variants
    pa = 1 - qa
    pb = 1 - qb

    # work around zero frequencies which cause problems when calculating logs
    pa[pa == 0] = np.exp(-250)
    qa[qa == 0] = np.exp(-250)
    pb[pb == 0] = np.exp(-250)
    qb[qb == 0] = np.exp(-250)

    # calculate likelihoods
    logpa = np.log(pa)
    logqa = np.log(qa)
    logpb = np.log(pb)
    logqb = np.log(qb)

    # set up likelihoods array
    n_models = 3
    n_gn_states = 3
    log_likelihoods = np.empty((n_variants, n_samples, n_models, n_gn_states),
                               dtype='f8')

    # probability of genotype (e.g., 0 = hom ref) given model (e.g., 0 = aa)
    log_likelihoods[:, :, 0, 0] = (2 * logpa)[:, np.newaxis]
    log_likelihoods[:, :, 1, 0] = (np.log(2) + logpa + logqa)[:, np.newaxis]
    log_likelihoods[:, :, 2, 0] = (2 * logqa)[:, np.newaxis]
    log_likelihoods[:, :, 0, 1] = (logpa + logpb)[:, np.newaxis]
    log_likelihoods[:, :, 1, 1] = (np.logaddexp(logpa + logqb,
                                                logqa + logpb)[:, np.newaxis])
    log_likelihoods[:, :, 2, 1] = (logqa + logqb)[:, np.newaxis]
    log_likelihoods[:, :, 0, 2] = (2 * logpb)[:, np.newaxis]
    log_likelihoods[:, :, 1, 2] = (np.log(2) + logpb + logqb)[:, np.newaxis]
    log_likelihoods[:, :, 2, 2] = (2 * logqb)[:, np.newaxis]

    # transform genotypes for convenience
    gn = anhima.gt.as_012(genotypes)

    # calculate actual model likelihoods for each genotype call
    model_likelihoods = np.empty((n_variants, n_samples, n_models), dtype='f8')
    model_likelihoods.fill(-250)
    for model in 0, 1, 2:
        for gn_state in 0, 1, 2:
            model_likelihoods[:, :, model][gn == gn_state] = \
                log_likelihoods[:, :, model, gn_state][gn == gn_state]

    # optionally combine likelihoods in a moving window
    if filter_size:
        model_likelihoods = np.apply_along_axis(np.convolve,
                                                0,
                                                model_likelihoods,
                                                np.ones((filter_size,)))
        # remove edges
        model_likelihoods = \
            model_likelihoods[filter_size//2:-1*(filter_size//2), ...]

    # predict ancestry as model with highest likelihood
    ancestry = np.argmax(model_likelihoods, axis=2)

    # calculate confidence by comparing first and second most likely models
    model_likelihoods.sort(axis=2)
    confidence = model_likelihoods[:, :, 2] - model_likelihoods[:, :, 1]

    # recind prediction where confidence is zero (models are equally likely)
    ancestry[confidence == 0] = -1

    return ancestry, confidence



# -*- coding: utf-8 -*-
"""
Doubleton sharing, a.k.a., analysis of f2 variants.

See also the examples at:

- http://nbviewer.ipython.org/github/alimanfoo/anhima/blob/master/examples/f2.ipynb

"""  # noqa


from __future__ import division, print_function, absolute_import


# standard library dependencies
import itertools


# third party dependencies
import numpy as np
import matplotlib.pyplot as plt
import scipy.special


def count_shared_doubletons(subpops_ac):
    """Count subpopulation pairs sharing doubletons (where one allele is
    observed in each subpopulation).

    Parameters
    ----------

    subpops_ac : array_like, int
        An array of shape (n_variants, n_subpops) holding alternate allele
        counts for each subpopulation.

    Returns
    -------

    counts : ndarray, int or float
        A square matrix of shape (n_subpops, n_subpops) where the array
        element at index (i, j) holds the count of shared doubletons between
        the ith and jth subpopulations.

    See Also
    --------

    normalise_doubleton_counts, plot_shared_doubletons,
    plot_total_doubletons, plot_f2_fig

    """

    # check input
    subpops_ac = np.asarray(subpops_ac)
    assert subpops_ac.ndim == 2

    # find doubletons in the total population
    is_total_doubleton = np.sum(subpops_ac, axis=1) == 2
    subpops_ac_doubletons = np.compress(is_total_doubleton, subpops_ac, axis=0)

    # count subpopulaton pairs sharing doubletons
    n_subpops = subpops_ac.shape[1]
    counts = np.zeros((n_subpops, n_subpops), dtype=np.int)
    for i in range(n_subpops):
        for j in range(i, n_subpops):
            if i == j:
                # count cases where doubleton is private to a subpopulation
                n = np.count_nonzero(subpops_ac_doubletons[:, i] == 2)
            else:
                # count cases where doubleton is shared between two
                # subpopulations
                n = np.count_nonzero((subpops_ac_doubletons[:, i] == 1)
                                     & (subpops_ac_doubletons[:, j] == 1))
            # fill both upper and lower triangles
            counts[i, j] = n
            counts[j, i] = n

    return counts


def normalise_doubleton_counts(counts, n_samples, ploidy=2):
    """Normalise doubleton counts by dividing by the number of distinct pairs of
    haplotypes in each population comparison.

    Parameters
    ----------

    counts : array_like, ints
        A square matrix of shape (n_subpops, n_subpops) where the array
        element at index (i, j) holds the count of shared doubletons between
        the ith and jth subpopulations.
    n_samples : int or sequence of ints
        The number of samples in each sub-population.
    ploidy : int, optional
        The sample ploidy.

    Returns
    -------

    normed_counts : ndarray, float
        Normalised counts of shared doubletons.

    See Also
    --------

    count_shared_doubletons

    Notes
    -----

    This function corrects for the fact that there are fewer pairs of
    haplotypes when looking for doubletons within a single subpopulation of
    size n than there are when comparing two different subpopulations of size
    n.

    This function may also help to correct for the case where the number of
    samples from each subpopulation is not equal. However, note that if this
    is the case then there may still also be some bias in how doubletons have
    been ascertained.

    """

    # check inputs
    counts = np.asarray(counts)
    assert counts.ndim == 2
    assert counts.shape[0] == counts.shape[1]
    n_subpops = counts.shape[0]

    # deal with polymorphic input
    if isinstance(n_samples, int):
        n_samples = [n_samples] * n_subpops
    n_samples = np.asarray(n_samples)

    # normalise counts
    normed_counts = np.zeros((n_subpops, n_subpops), dtype=np.float)
    for i in range(n_subpops):
        for j in range(i, n_subpops):
            if i == j:
                # number of distinct pairs of haplotypes within a
                # subpopulation = number of haplotypes choose 2
                n_haplotypes = n_samples[i] * ploidy
                n_pairs = scipy.special.comb(n_haplotypes, 2)
            else:
                # number of distinct pairs of haplotypes between
                # subpopulations
                n_pairs = (n_samples[i] * ploidy) * (n_samples[j] * ploidy)
            # fill upper and lower triangles
            normed_counts[i, j] = counts[i, j] / n_pairs
            normed_counts[j, i] = counts[j, i] / n_pairs

    return normed_counts


def plot_shared_doubletons(counts, subpop_labels=None,
                           subpop_colors='bgrcmyk', axs=None,
                           figsize_factor=1, ylim=None, relative=False,
                           flip=False):
    """Plot counts of doubleton sharing between subpopulations as a bar chart.

    Parameters
    ----------

    counts : array_like, ints
        A square matrix of shape (n_subpops, n_subpops) where the array
        element at index (i, j) holds the count of shared doubletons between
        the ith and jth subpopulations.
    subpop_labels : sequence of strings, optional
        Labels for the subpopulations.
    subpop_colors : sequence of colors, optional
        Colors for the subpopulations.
    axs : sequence of axes, optional
        The axes to use. If not provided, a new figure will be created.
    figsize_factor : float, optional
        Figure size in inches per subpopulation. Only used if `axs` is None.
    ylim : pair of ints or floats, optional
        Limits for the Y axes of all subplots.
    relative : bool, optional
        If True, normalise counts by dividing by the sum along each row.
    flip : bool, optional
        If True, invert the Y axis.

    Returns
    -------

    axs : sequence of axes
        The axes on which the plot was drawn.

    See Also
    --------

    count_shared_doubletons, plot_total_doubletons, plot_f2_fig

    """

    # check inputs
    counts = np.asarray(counts)
    assert counts.ndim == 2
    assert counts.shape[0] == counts.shape[1]
    n_subpops = counts.shape[0]

    # setup axes
    if axs is None:
        x = n_subpops * figsize_factor
        fig = plt.figure(figsize=(x, x))
        axs = [fig.add_subplot(n_subpops, 1, i+1) for i in range(n_subpops)]

    # ensure we have enough colors
    colors = list(itertools.islice(itertools.cycle(subpop_colors), n_subpops))

    # ensure we have subpopulation labels
    if subpop_labels is None:
        subpop_labels = range(n_subpops)

    # normalise by row
    if relative:
        # compute sum along each row
        row_sum = np.sum(counts, axis=1)
        # make sure we get the broadcasting right
        counts = counts / row_sum[:, np.newaxis]

    # determine global Y limits
    if ylim is None:
        ylim = (0, np.amax(counts))

    # plot main bar
    for i, color in zip(range(n_subpops), colors):

        if flip:
            # plot from the top down
            ax = axs[i]
        else:
            # plot from the bottom up
            ax = axs[n_subpops-i-1]

        # select data to plot
        data = counts[i, :]

        # make a bar
        ax.bar(range(n_subpops), data, width=1, color=colors)

        # tidy up
        ax.set_ylim(*ylim)
        ax.set_ylabel(subpop_labels[i], rotation=0, ha='right', va='center',
                      color=color)
        ax.tick_params(length=0)
        ax.set_yticks([])
        if (flip and i > 0) or (not flip and i < n_subpops - 1):
            ax.set_xticks([])
        else:
            ax.xaxis.tick_top()
            ax.set_xticks(np.arange(n_subpops) + .5)
            ax.set_xticklabels(subpop_labels, rotation=90)
        for s in 'top', 'left', 'right':
            ax.spines[s].set_visible(False)

    return axs


def plot_total_doubletons(counts, subpop_labels=None,
                          width=.8, orientation='vertical', n_samples=None,
                          ax=None, bar_kwargs=None):
    """Plot total counts of doubletons per subpopulations as a bar chart.

    Parameters
    ----------

    counts : array_like, ints
        A square matrix of shape (n_subpops, n_subpops) where the array
        element at index (i, j) holds the count of shared doubletons between
        the ith and jth subpopulations.
    subpop_labels : sequence of strings, optional
        Labels for the subpopulations.
    width : float, optional
        The relative width of each bar.
    orientation : {'vertical', 'horizontal'}
        The bar orientation.
    n_samples : int or sequence of ints
        The number of samples in each sub-population.
    ax : axes, optional
        The axes on which to plot. If not provided, a new figure will be
        created.
    bar_kwargs : dict, optional
        Keyword arguments passed through to ax.bar().

    Returns
    -------

    ax : axes
        The axes on which the plot was drawn.

    See Also
    --------

    count_shared_doubletons, plot_shared_doubletons, plot_total_doubletons,
    plot_f2_fig

    """

    # check inputs
    counts = np.asarray(counts)
    assert counts.ndim == 2
    assert counts.shape[0] == counts.shape[1]
    n_subpops = counts.shape[0]

    # setup axes
    if ax is None:
        fig, ax = plt.subplots()

    # sum rows
    y = np.sum(counts, axis=1)

    # normalise total counts by number of samples
    if n_samples is not None:
        # deal with polymorphic input
        if isinstance(n_samples, int):
            n_samples = [n_samples] * n_subpops
        n_samples = np.asarray(n_samples)
        y = y / n_samples

    # plot bar
    x = np.arange(n_subpops) + .5
    if bar_kwargs is None:
        bar_kwargs = dict()
    bar_kwargs.setdefault('color', 'gray')
    bar_kwargs.setdefault('align', 'center')
    bar_kwargs.setdefault('linewidth', 0)
    if orientation == 'vertical':
        ax.bar(x, y, width, **bar_kwargs)
    else:
        ax.barh(x, y, width, **bar_kwargs)

    # ensure we have subpopulation labels
    if subpop_labels is None:
        subpop_labels = range(n_subpops)

    # tidy up
    ax.tick_params(length=0)
    if orientation == 'vertical':
        for s in 'top', 'right', 'left':
            ax.spines[s].set_visible(False)
        ax.set_xticks(range(n_subpops))
        ax.set_xticklabels(subpop_labels, rotation=90)
        ax.set_yticks([])
        ax.xaxis.tick_bottom()
        ax.set_ylabel('doubletons')
        ax.set_xlim(0, n_subpops)
    else:
        for s in 'top', 'right', 'bottom':
            ax.spines[s].set_visible(False)
        ax.set_yticks(range(n_subpops))
        ax.set_yticklabels(subpop_labels, rotation=0)
        ax.set_xticks([])
        ax.yaxis.tick_left()
        ax.set_xlabel('doubletons')
        ax.xaxis.set_label_position('top')
        ax.set_ylim(0, n_subpops)

    return ax


def plot_f2_fig(counts, subpop_labels=None, subpop_colors='bgrcmyk', fig=None,
                figsize_factor=1, relative=False, normed=False, n_samples=None,
                ploidy=2):
    """Plot a combined figure of shared doubleton counts and total counts per
    subpopulation.

    Parameters
    ----------

    counts : array_like, ints
        A square matrix of shape (n_subpops, n_subpops) where the array
        element at index (i, j) holds the count of shared doubletons between
        the ith and jth subpopulations.
    subpop_labels : sequence of strings, optional
        Labels for the subpopulations.
    subpop_colors : sequence of colors, optional
        Colors for the subpopulations.
    fig : figure, optional
        The figure to use. If not provided, a new figure will be created.
    figsize_factor : float, optional
        Figure size in inches per subpopulation. Only used if `fig` is None.
    relative : bool, optional
        If True, plot counts relative to the sum along each row.
    normed : bool, optional
        If True, normalise counts by dividing by the number of possible
        pairs of haplotypes.
    n_samples : int or sequence of ints
        The number of samples in each sub-population.
    ploidy : int, optional
        The sample ploidy. (Only relevant if `normed` is True.)

    Returns
    -------

    fig : figure
        The figure on which the plot was drawn.

    See Also
    --------

    count_shared_doubletons, plot_shared_doubletons, plot_total_doubletons

    """

    # check inputs
    counts = np.asarray(counts)
    assert counts.ndim == 2
    assert counts.shape[0] == counts.shape[1]
    n_subpops = counts.shape[0]

    # setup figure
    if fig is None:
        width = (n_subpops + 1) * figsize_factor
        height = n_subpops * figsize_factor
        fig = plt.figure(figsize=(width, height))

    # plot main bar
    main_axs = [plt.subplot2grid((n_subpops, n_subpops+1), (i, 0), rowspan=1,
                                 colspan=n_subpops)
                for i in range(n_subpops)]
    if normed:
        main_counts = normalise_doubleton_counts(counts,
                                                 n_samples=n_samples,
                                                 ploidy=ploidy)
    else:
        main_counts = counts
    plot_shared_doubletons(main_counts,
                           subpop_labels=subpop_labels,
                           subpop_colors=subpop_colors,
                           relative=relative,
                           axs=main_axs)

    # plot totals bar
    tot_ax = plt.subplot2grid((n_subpops, n_subpops+1), (0, n_subpops),
                              rowspan=n_subpops, colspan=1)
    plot_total_doubletons(counts, subpop_labels=subpop_labels,
                          n_samples=n_samples, orientation='horizontal',
                          ax=tot_ax)
    tot_ax.set_yticks([])

    return fig



# -*- coding: utf-8 -*-
# flake8: noqa
from __future__ import absolute_import, print_function, division


import sys
import itertools


PY2 = sys.version_info[0] == 2


if PY2:
    range = xrange
    map = itertools.imap
    string_types = basestring,
    zip = itertools.izip
else:
    range = range
    map = map
    string_types = str,
    zip = zip


def force_bytes(o):
    if PY2:
        return o
    elif isinstance(o, str):
        return o.encode('ascii')
    elif isinstance(o, list):
        return [force_bytes(x) for x in o]
    elif isinstance(o, tuple):
        return tuple(force_bytes(x) for x in o)
    else:
        return o



# -*- coding: utf-8 -*-
"""
Site frequency spectra.

See also the examples at:

- http://nbviewer.ipython.org/github/alimanfoo/anhima/blob/master/examples/sf.ipynb

"""  # noqa


from __future__ import division, print_function, absolute_import


# third party dependencies
import numpy as np
import matplotlib.pyplot as plt
import scipy.stats


def site_frequency_spectrum(derived_ac):
    """Calculate the site frequency spectrum, given derived allele counts for a
    set of biallelic variant sites.

    Parameters
    ----------

    derived_ac : array_like, int
        A 1-dimensional array of shape (n_variants,) where each array
        element holds the count of derived alleles found for a single variant
        across some set of samples.

    Returns
    -------

    sfs : ndarray, int
        An array of integers where the value of the kth element is the
        number of variant sites with k derived alleles.

    See Also
    --------

    site_frequency_spectrum_scaled, site_frequency_spectrum_folded,
    site_frequency_spectrum_folded_scaled, plot_site_frequency_spectrum

    """

    # check input
    derived_ac = np.asarray(derived_ac)
    assert derived_ac.ndim == 1

    # calculate frequency spectrum
    sfs = np.bincount(derived_ac)

    return sfs


def site_frequency_spectrum_folded(biallelic_ac):
    """Calculate the folded site frequency spectrum, given reference and
    alternate allele counts for a set of biallelic variants.

    Parameters
    ----------

    biallelic_ac : array_like int
        A 2-dimensional array of shape (n_variants, 2), where each row
        holds the reference and alternate allele counts for a single
        biallelic variant across some set of samples.

    Returns
    -------

    sfs_folded : ndarray, int
        An array of integers where the value of the kth element is the
        number of variant sites with k observations of the minor allele.

    See Also
    --------

    site_frequency_spectrum, site_frequency_spectrum_scaled,
    site_frequency_spectrum_folded_scaled, plot_site_frequency_spectrum

    """

    # check input
    biallelic_ac = np.asarray(biallelic_ac)
    assert biallelic_ac.ndim == 2
    assert biallelic_ac.shape[1] == 2

    # calculate minor allele counts
    minor_ac = np.amin(biallelic_ac, axis=1)

    # calculate frequency spectrum
    sfs_folded = np.bincount(minor_ac)

    return sfs_folded


def site_frequency_spectrum_scaled(derived_ac):
    """Calculate the site frequency spectrum, scaled such that a constant value
    is expected across the spectrum for neutral variation and a population at
    constant size.

    Parameters
    ----------

    derived_ac : array_like, int
        A 1-dimensional array of shape (n_variants,) where each array
        element holds the count of derived alleles found for a single variant
        across some set of samples.

    Returns
    -------

    sfs_scaled : ndarray, int
        An array of integers where the value of the kth element is the
        number of variant sites with k derived alleles, multiplied by k.

    Notes
    -----

    Under neutrality and constant population size, site frequency
    is expected to be constant across the spectrum, and to approximate
    the value of the population-scaled mutation rate theta.

    See Also
    --------

    site_frequency_spectrum, site_frequency_spectrum_folded,
    site_frequency_spectrum_folded_scaled, plot_site_frequency_spectrum

    """

    # calculate frequency spectrum
    sfs = site_frequency_spectrum(derived_ac)

    # scaling
    k = np.arange(sfs.size)
    sfs_scaled = sfs * k

    return sfs_scaled


def site_frequency_spectrum_folded_scaled(biallelic_ac, m=None):
    """Calculate the folded site frequency spectrum, scaled such that a
    constant value is expected across the spectrum for neutral variation and
    a population at constant size.

    Parameters
    ----------

    biallelic_ac : array_like int
        A 2-dimensional array of shape (n_variants, 2), where each row
        holds the reference and alternate allele counts for a single
        biallelic variant across some set of samples.
    m : int, optional
        The total number of alleles observed at each variant site. Equal to
        the number of samples multiplied by the ploidy. If not provided,
        will be inferred to be the maximum value of the sum of reference and
        alternate allele counts present in `biallelic_ac`.

    Returns
    -------

    sfs_folded_scaled : ndarray, int
        An array of integers where the value of the kth element is the
        number of variant sites with k observations of the minor allele,
        multiplied by the scaling factor (k * (m - k) / m).

    Notes
    -----

    Under neutrality and constant population size, site frequency
    is expected to be constant across the spectrum, and to approximate
    the value of the population-scaled mutation rate theta.

    This function is useful where the ancestral and derived status of alleles
    is unknown.

    See Also
    --------

    site_frequency_spectrum, site_frequency_spectrum_scaled,
    site_frequency_spectrum_folded, plot_site_frequency_spectrum

    """

    # calculate the folded site frequency spectrum
    sfs_folded = site_frequency_spectrum_folded(biallelic_ac)

    # determine the total number of alleles per variant
    if m is None:
        m = np.amax(np.sum(biallelic_ac, axis=1))

    # scaling
    k = np.arange(sfs_folded.size)
    sfs_folded_scaled = sfs_folded * k * (m - k) / m

    return sfs_folded_scaled


def plot_site_frequency_spectrum(sfs, bins=None, m=None,
                                 clip_endpoints=True, ax=None, label=None,
                                 plot_kwargs=None):
    """Plot a site frequency spectrum.

    Parameters
    ----------

    sfs : array_like, int
        Site frequency spectrum. Can be folded or unfolded, scaled or
        unscaled.
    bins : int or sequence of ints, optional
        Number of bins or bin edges to aggregate frequencies. If not given,
        no binning will be applied.
    m : int, optional
        The total number of alleles observed at each variant site. Equal to
        the number of samples multiplied by the ploidy. If given, will be
        used to scale the X axis as allele frequency instead of allele count.
        used to scale the X axis as allele frequency instead of allele count.
    clip_endpoints : bool, optional
        If True, remove the first and last values from the site frequency
        spectrum.
    ax : axes, optional
        The axes on which to plot. If not given, a new figure will be created.
    label : string, optional
        Label for this data series.
    plot_kwargs : dict, optional
        Passed through to ax.plot().

    Returns
    -------

    ax : axes
        The axes on which the plot was drawn.

    See Also
    --------

    site_frequency_spectrum, site_frequency_spectrum_folded,
    site_frequency_spectrum_scaled, site_frequency_spectrum_folded_scaled

    """

    if ax is None:
        fig, ax = plt.subplots()

    if bins is None:
        # no binning
        if clip_endpoints:
            x = np.arange(1, sfs.size-1)
            y = sfs[1:-1]
        else:
            x = np.arange(sfs.size)
            y = sfs

    else:
        # bin the frequencies
        if clip_endpoints:
            y, b, _ = scipy.stats.binned_statistic(np.arange(1, sfs.size-1),
                                                   values=sfs[1:-1],
                                                   bins=bins,
                                                   statistic='mean')
        else:
            y, b, _ = scipy.stats.binned_statistic(np.arange(sfs.size),
                                                   values=sfs,
                                                   bins=bins,
                                                   statistic='mean')
        # use bin midpoints for plotting
        x = (b[:-1] + b[1:]) / 2

    if m is not None:
        # convert allele counts to allele frequencies
        x = x / m
        ax.set_xlabel('allele frequency')
    else:
        ax.set_xlabel('allele count')

    # plotting
    if plot_kwargs is None:
        plot_kwargs = dict()
    ax.plot(x, y, label=label, **plot_kwargs)

    # tidy up
    ax.set_ylabel('site frequency')

    return ax



# -*- coding: utf-8 -*-
from __future__ import division, print_function, absolute_import


import unittest
import numpy as np
from anhima.loc import locate_position, locate_interval, locate_positions, \
    locate_intervals


class TestLocatePositions(unittest.TestCase):

    def test_locate_position(self):
        pos = np.array([3, 6, 11])
        self.assertEqual(0, locate_position(pos, 3))
        self.assertEqual(1, locate_position(pos, 6))
        self.assertEqual(2, locate_position(pos, 11))
        self.assertIsNone(locate_position(pos, 1))
        self.assertIsNone(locate_position(pos, 12))

    def test_locate_positions(self):
        pos1 = np.array([3, 6, 11, 20, 35])
        pos2 = np.array([4, 6, 20, 39])
        expect_cond1 = np.array([False, True, False, True, False])
        expect_cond2 = np.array([False, True, True, False])
        cond1, cond2 = locate_positions(pos1, pos2)
        assert np.array_equal(expect_cond1, cond1)
        assert np.array_equal(expect_cond2, cond2)

    def test_locate_interval(self):
        pos = np.array([3, 6, 11, 20, 35])
        self.assertEqual(slice(0, 5), locate_interval(pos, 2, 37))
        self.assertEqual(slice(1, 5), locate_interval(pos, 4, 37))
        self.assertEqual(slice(0, 4), locate_interval(pos, 2, 32))
        self.assertEqual(slice(1, 4), locate_interval(pos, 4, 32))
        self.assertEqual(slice(1, 3), locate_interval(pos, 4, 19))
        self.assertEqual(slice(2, 4), locate_interval(pos, 7, 32))
        self.assertEqual(slice(2, 3), locate_interval(pos, 7, 19))
        self.assertEqual(slice(3, 3), locate_interval(pos, 17, 19))
        self.assertEqual(slice(0, 0), locate_interval(pos, 0, 0))
        self.assertEqual(slice(5, 5), locate_interval(pos, 1000, 2000))

    def test_locate_intervals(self):
        pos = np.array([3, 6, 11, 20, 35])
        intervals = np.array([[0, 2], [6, 17], [12, 15], [31, 35], [100, 120]])
        expect_cond1 = np.array([False, True, True, False, True])
        expect_cond2 = np.array([False, True, False, True, False])
        cond1, cond2 = locate_intervals(pos, intervals[:, 0], intervals[:, 1])
        assert np.array_equal(expect_cond1, cond1)
        assert np.array_equal(expect_cond2, cond2)
        assert np.array_equal([6, 11, 35], pos[cond1])
        assert np.array_equal([[6, 17], [31, 35]], intervals[cond2])



# -*- coding: utf-8 -*-
"""
Utility functions for multidimensional scaling.

R must be installed, and the Python package ``rpy2`` must be installed, e.g.::

    $ apt-get install r-base
    $ pip install rpy2

See also the examples at:

- http://nbviewer.ipython.org/github/alimanfoo/anhima/blob/master/examples/mds.ipynb

"""  # noqa


from __future__ import division, print_function, absolute_import


# third party dependencies
import numpy as np
import matplotlib.pyplot as plt
import sklearn.manifold


_r_initialised = False
rpy2 = None
ro = None
r = None


def _init_r():
    """Private function to initialise R, only executed when needed.

    """

    global _r_initialised
    global rpy2
    global ro
    global r

    if not _r_initialised:
        import rpy2  # noqa
        import rpy2.robjects as ro  # noqa
        from rpy2.robjects import r  # noqa
        import rpy2.robjects.numpy2ri as numpy2ri
        numpy2ri.activate()
        _r_initialised = True


def smacof(dist_square, **kwargs):
    """
    Multidimensional scaling using the SMACOF (Scaling by Majorizing a
    Complicated Function) algorithm.

    Parameters
    ----------

    dist_square : array_like, shape (n_samples, n_samples)
        A distance matrix in square form.
    kwargs : additional keyword arguments
        Additional keyword arguments are passed through to
        :func:`sklearn.manifold.MDS`.

    Returns
    -------

    coords : ndarray, shape (n_samples, n_components)
        An array whose rows give the coordinates of the points chosen to
        represent the dissimilarities.

    See Also
    --------

    anhima.dist.pairwise_distance, anhima.mds.classical, sklearn.manifold.MDS,
    anhima.pca.pca

    """

    # normalise inputs
    dist_square = np.asarray(dist_square)
    assert dist_square.ndim == 2
    assert dist_square.shape[0] == dist_square.shape[1]

    # setup model
    model = sklearn.manifold.MDS(dissimilarity='precomputed',
                                 **kwargs)

    # fit model and get transformed coordinates
    coords = model.fit(dist_square).embedding_

    return coords


def classical(dist_square, k=2):
    """
    Classical multidimensional scaling via the R ``cmdscale`` function.

    Parameters
    ----------

    dist_square : array_like, shape (n_samples, n_samples)
        A distance matrix in square form.
    k : integer, optional
        The maximum dimension of the space which the data are to be represented
        in; must be in {1, 2, ..., n-1}.

    Returns
    -------

    coords : ndarray, shape (n_samples, k)
        An array whose rows give the coordinates of the points chosen to
        represent the dissimilarities.

    See Also
    --------

    anhima.dist.pairwise_distance, anhima.mds.smacof, anhima.pca.pca

    """

    # setup R
    _init_r()

    # normalise inputs
    dist_square = np.asarray(dist_square)
    assert dist_square.ndim == 2
    assert dist_square.shape[0] == dist_square.shape[1]

    # convert distance matrix to R
    m = ro.vectors.Matrix(dist_square)

    # apply MDS
    coords = r['cmdscale'](m, k=k)

    return np.asarray(coords)


def plot_coords(coords, dimx=1, dimy=2, ax=None, colors='b', sizes=20,
                labels=None, scatter_kwargs=None, annotate_kwargs=None):
    """Scatter plot of transformed coordinates from multidimensional scaling.

    Parameters
    ----------

    coords : ndarray, shape (`n_samples`, `n_components`)
        The transformed coordinates.
    dimx : int, optional
        The dimension to plot on the X axis. N.B., this is
        one-based, so `1` is the first dimension, `2` is the second
        dimension, etc.
    dimy : int, optional
        The dimension to plot on the Y axis. N.B., this is
        one-based, so `1` is the first dimension, `2` is the second
        dimension, etc.
    ax : axes, optional
        The axes on which to draw. If not provided, a new figure will be
        created.
    colors : color or sequence of color, optional
        Can be a single color format string, or a sequence of color
        specifications of length `n_samples`.
    sizes : scalar or array_like, shape (`n_samples`), optional
        Size in points^2.
    labels : sequence of strings
        If provided, will be used to label points in the plot.
    scatter_kwargs : dict-like
        Additional keyword arguments passed through to `plt.scatter`.
    annotate_kwargs : dict-like
        Additional keyword arguments passed through to `plt.annotate` when
        labelling points.


    Returns
    -------

    ax : axes
        The axes on which the plot was drawn.

    See Also
    --------

    anhima.mds.smacof, anhima.mds.classical

    """

    # normalise inputs
    coords = np.asarray(coords)
    assert coords.ndim == 2

    # set up axes
    if ax is None:
        # make a square figure
        x = plt.rcParams['figure.figsize'][0]
        fig, ax = plt.subplots(figsize=(x, x))

    # obtain X and Y data, N.B., `pcx` and `pcy` are 1-based
    x = coords[:, dimx-1]
    y = coords[:, dimy-1]

    # plot points
    if scatter_kwargs is None:
        scatter_kwargs = dict()
    ax.scatter(x, y, c=colors, s=sizes, **scatter_kwargs)

    # label points
    if labels is not None:
        if annotate_kwargs is None:
            annotate_kwargs = dict()
        annotate_kwargs.setdefault('xycoords', 'data')
        annotate_kwargs.setdefault('xytext', (3, 3))
        annotate_kwargs.setdefault('textcoords', 'offset points')
        for l, lx, ly in zip(labels, x, y):
            if l is not None:
                ax.annotate(str(l), xy=(lx, ly), **annotate_kwargs)

    # tidy up
    ax.set_xlabel('dimension %s' % dimx)
    ax.set_ylabel('dimension %s' % dimy)

    return ax



# -*- coding: utf-8 -*-
from __future__ import division, print_function, absolute_import


import unittest
import numpy as np
import os
import anhima.sim
import anhima.io
import scipy.stats
import tempfile
import logging


logger = logging.getLogger(__name__)
debug = logger.debug


class TestTped(unittest.TestCase):

    def setUp(self):

        n_variants = 1000
        self.ref = np.random.choice(['A', 'C', 'G', 'T'], n_variants)
        self.alt = np.random.choice(['A', 'C', 'G', 'T'], n_variants)
        self.pos = np.random.choice(range(n_variants*100), n_variants, False)
        self.pos.sort()

        # simulate genotypes
        self.n_samples = 100
        ploidy = 2
        af_dist = scipy.stats.beta(a=.4, b=.6)
        p_missing = .1
        self.genotypes = anhima.sim.simulate_biallelic_genotypes(
            n_variants, self.n_samples, af_dist, p_missing, ploidy)

        self.n_variants = n_variants

    def test_file_created(self):

        path = tempfile.NamedTemporaryFile(delete=False)
        debug(path.name)
        anhima.io.save_tped(path.name, self.genotypes, self.ref, self.alt,
                            self.pos)
        self.assertTrue(os.path.isfile(path.name))

        # read in file
        with open(path.name) as f:
            content = f.readlines()

        self.assertEqual(self.n_variants, len(content))

        # test that content is good...by taking first line
        line = content[0].split("\t")
        self.assertEquals(self.n_samples + 4, len(line))
        self.assertEquals('_'.join(line[0:4]),
                          '_'.join(['0', 'snp' + str(self.pos[0]),
                                    '0.0', str(self.pos[0])]))
    # test that we can load from hdf5 and create tped

    # test that we can write to hdf5 ok



# -*- coding: utf-8 -*-
# flake8: noqa
from __future__ import division, print_function, absolute_import


__version__ = '0.11.1'


# make sub-packages available without explicit import
from . import af
from . import dist
from . import f2
from . import gt
from . import h5
from . import io
from . import ld
from . import loc
from . import mds
from . import pca
from . import ped
from . import sf
from . import sim
from . import tree
from . import util



# -*- coding: utf-8 -*-
"""
Utilities for working with related individuals (crosses, families, etc.).

See also the examples at:

- http://nbviewer.ipython.org/github/alimanfoo/anhima/blob/master/examples/ped.ipynb

"""  # noqa


from __future__ import division, print_function, absolute_import


# third party dependencies
import numpy as np
import numexpr as ne
import pandas

# internal dependencies
import anhima.gt

# constants to represent inheritance states
INHERIT_UNDETERMINED = 0
INHERIT_PARENT1 = 1
INHERIT_PARENT2 = 2
INHERIT_NONSEG_REF = 3
INHERIT_NONSEG_ALT = 4
INHERIT_NONPARENTAL = 5
INHERIT_PARENT_MISSING = 6
INHERIT_MISSING = 7
INHERITANCE_STATES = range(8)
INHERITANCE_LABELS = ('undetermined', 'parent1', 'parent2', 'non-seg ref',
                      'non-seg alt', 'non-parental', 'parent missing',
                      'missing')


def diploid_inheritance(parent_diplotype, gamete_haplotypes):
    """
    Determine the transmission of parental alleles to a set of gametes.

    Parameters
    ----------

    parent_diplotype : array_like, shape (n_variants, 2)
        An array of phased genotypes for a single diploid individual, where
        each element of the array is an integer corresponding to an allele
        index (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).
    gamete_haplotypes : array_like, shape (n_variants, n_gametes)
        An array of haplotypes for a set of gametes derived from the given
        parent, where each element of the array is an integer corresponding
        to an allele index (-1 = missing, 0 = reference allele, 1 = first
        alternate allele, 2 = second alternate allele, etc.).

    Returns
    -------

    inheritance : ndarray, uint8, shape (n_variants, n_gametes)
        An array of integers coding the allelic inheritance, where 1 =
        inheritance from first parental haplotype, 2 = inheritance from second
        parental haplotype, 3 = inheritance of reference allele from parent
        that is homozygous for the reference allele, 4 = inheritance of
        alternate allele from parent that is homozygous for the alternate
        allele, 5 = non-parental allele, 6 = parental genotype is missing,
        7 = gamete allele is missing.

    """

    # normalise inputs
    parent_diplotype = np.asarray(parent_diplotype)
    assert parent_diplotype.ndim == 2
    assert parent_diplotype.shape[1] == 2
    gamete_haplotypes = np.asarray(gamete_haplotypes)
    assert gamete_haplotypes.ndim == 2

    # convenience variables
    parent1 = parent_diplotype[:, 0, np.newaxis]  # noqa
    parent2 = parent_diplotype[:, 1, np.newaxis]  # noqa
    gamete_is_missing = gamete_haplotypes < 0
    parent_is_missing = np.any(parent_diplotype < 0, axis=1)
    parent_is_hom_ref = anhima.gt.is_hom_ref(parent_diplotype)[:, np.newaxis]  # noqa
    parent_is_het = anhima.gt.is_het(parent_diplotype)[:, np.newaxis]  # noqa
    parent_is_hom_alt = anhima.gt.is_hom_alt(parent_diplotype)[:, np.newaxis]  # noqa

    # need this for broadcasting, but also need to retain original for later
    parent_is_missing_bc = parent_is_missing[:, np.newaxis]  # noqa

    # N.B., use numexpr below where possible to avoid temporary arrays

    # utility variable, identify allele calls where inheritance can be
    # determined
    callable = ne.evaluate('~gamete_is_missing & ~parent_is_missing_bc')  # noqa
    callable_seg = ne.evaluate('callable & parent_is_het')  # noqa

    # main inheritance states
    inherit_parent1 = ne.evaluate(
        'callable_seg & (gamete_haplotypes == parent1)'
    )
    inherit_parent2 = ne.evaluate(
        'callable_seg & (gamete_haplotypes == parent2)'
    )
    nonseg_ref = ne.evaluate(
        'callable & parent_is_hom_ref & (gamete_haplotypes == parent1)'
    )
    nonseg_alt = ne.evaluate(
        'callable & parent_is_hom_alt & (gamete_haplotypes == parent1)'
    )
    nonparental = ne.evaluate(
        'callable & (gamete_haplotypes != parent1)'
        ' & (gamete_haplotypes != parent2)'
    )

    # record inheritance states
    # N.B., order in which these are set matters
    inheritance = np.zeros_like(gamete_haplotypes, dtype='u1')
    inheritance[inherit_parent1] = INHERIT_PARENT1
    inheritance[inherit_parent2] = INHERIT_PARENT2
    inheritance[nonseg_ref] = INHERIT_NONSEG_REF
    inheritance[nonseg_alt] = INHERIT_NONSEG_ALT
    inheritance[nonparental] = INHERIT_NONPARENTAL
    inheritance[parent_is_missing] = INHERIT_PARENT_MISSING
    inheritance[gamete_is_missing] = INHERIT_MISSING

    return inheritance


def diploid_mendelian_error_biallelic(parental_genotypes, progeny_genotypes):
    """Implementation of function to find Mendelian errors optimised for
    biallelic variants.

    """

    # recode genotypes for convenience
    parental_genotypes_012 = anhima.gt.as_012(parental_genotypes)
    progeny_genotypes_012 = anhima.gt.as_012(progeny_genotypes)

    # convenience variables
    p1 = parental_genotypes_012[:, 0, np.newaxis]  # noqa
    p2 = parental_genotypes_012[:, 1, np.newaxis]  # noqa
    o = progeny_genotypes_012  # noqa

    # enumerate all possible combinations of Mendel error genotypes
    ex = '((p1 == 0) & (p2 == 0) & (o == 1))' \
        ' + ((p1 == 0) & (p2 == 0) & (o == 2)) * 2' \
        ' + ((p1 == 2) & (p2 == 2) & (o == 1))' \
        ' + ((p1 == 2) & (p2 == 2) & (o == 0)) * 2' \
        ' + ((p1 == 0) & (p2 == 2) & (o == 0))' \
        ' + ((p1 == 0) & (p2 == 2) & (o == 2))' \
        ' + ((p1 == 2) & (p2 == 0) & (o == 0))' \
        ' + ((p1 == 2) & (p2 == 0) & (o == 2))' \
        ' + ((p1 == 0) & (p2 == 1) & (o == 2))' \
        ' + ((p1 == 1) & (p2 == 0) & (o == 2))' \
        ' + ((p1 == 2) & (p2 == 1) & (o == 0))' \
        ' + ((p1 == 1) & (p2 == 2) & (o == 0))'
    errors = ne.evaluate(ex).astype('u1')

    return errors


def diploid_mendelian_error_multiallelic(parental_genotypes, progeny_genotypes,
                                         max_allele):
    """Implementation of function to find Mendelian errors generalised for
    multiallelic variants.

    """

    # transform genotypes into per-call allele counts
    alleles = range(max_allele + 1)
    p = anhima.gt.as_allele_counts(parental_genotypes, alleles=alleles)
    o = anhima.gt.as_allele_counts(progeny_genotypes, alleles=alleles)

    # detect nonparental and hemiparental inheritance by comparing allele
    # counts between parents and progeny
    ps = p.sum(axis=1)[:, np.newaxis]  # add allele counts for both parents
    ac_diff = (o - ps).astype('i1')
    ac_diff[ac_diff < 0] = 0
    # sum over all alleles
    errors = np.sum(ac_diff, axis=2).astype('u1')

    # detect uniparental inheritance by finding cases where no alleles are
    # shared between parents, then comparing progeny allele counts to each
    # parent
    pc1 = p[:, 0, np.newaxis, :]
    pc2 = p[:, 1, np.newaxis, :]
    # find variants where parents don't share any alleles
    is_shared_allele = (pc1 > 0) & (pc2 > 0)
    no_shared_alleles = ~np.any(is_shared_allele, axis=2)
    # find calls where progeny genotype is identical to one or the other parent
    errors[
        no_shared_alleles
        & (np.all(o == pc1, axis=2)
           | np.all(o == pc2, axis=2))
    ] = 1

    # retrofit where either or both parent has a missing call
    is_parent_missing = anhima.gt.is_missing(parental_genotypes)
    errors[np.any(is_parent_missing, axis=1)] = 0

    return errors


def diploid_mendelian_error(parental_genotypes, progeny_genotypes):
    """Find impossible genotypes according to Mendelian inheritance laws.

    Parameters
    ----------

    parental_genotypes : array_like, int
        An array of shape (n_variants, 2, 2) where each element of the array
        is an integer corresponding to an allele index (-1 = missing,
        0 = reference allele, 1 = first alternate allele, 2 = second
        alternate allele, etc.).
    progeny_genotypes : array_like, int
        An array of shape (n_variants, n_progeny, 2) where each element of the
        array is an integer corresponding to an allele index (-1 = missing,
        0 = reference allele, 1 = first alternate allele, 2 = second
        alternate allele, etc.).

    Returns
    -------

    errors : ndarray, uint8
        An array of shape (n_variants, n_progeny) where each element counts
        the number of non-Mendelian alleles in a progeny genotype call.

    See Also
    --------

    count_diploid_mendelian_error

    Notes
    -----

    Not applicable to polyploid genotype calls.

    Applicable to multiallelic variants.

    Assumes that genotypes are unphased.

    """

    # check inputs
    parental_genotypes = np.asarray(parental_genotypes)
    progeny_genotypes = np.asarray(progeny_genotypes)
    assert parental_genotypes.ndim == 3
    assert progeny_genotypes.ndim == 3

    # check the number of variants is equal in parents and progeny
    assert parental_genotypes.shape[0] == progeny_genotypes.shape[0]

    # check the number of parents
    assert parental_genotypes.shape[1] == 2

    # check the ploidy
    assert parental_genotypes.shape[2] == progeny_genotypes.shape[2] == 2

    # determine which implementation to use
    max_allele = max(np.amax(parental_genotypes), np.amax(progeny_genotypes))
    if max_allele < 2:
        errors = diploid_mendelian_error_biallelic(parental_genotypes,
                                                   progeny_genotypes)
    else:
        errors = diploid_mendelian_error_multiallelic(parental_genotypes,
                                                      progeny_genotypes,
                                                      max_allele)

    return errors


def count_diploid_mendelian_error(parental_genotypes,
                                  progeny_genotypes,
                                  axis=None):
    """Count impossible genotypes according to Mendelian inheritance laws,
    summed over all progeny genotypes, or summed along variants or samples.

    Parameters
    ----------

    parental_genotypes : array_like, int
        An array of shape (n_variants, 2, 2) where each element of the array
        is an integer corresponding to an allele index (-1 = missing,
        0 = reference allele, 1 = first alternate allele, 2 = second
        alternate allele, etc.).
    progeny_genotypes : array_like, int
        An array of shape (n_variants, n_progeny, 2) where each element of the
        array is an integer corresponding to an allele index (-1 = missing,
        0 = reference allele, 1 = first alternate allele, 2 = second
        alternate allele, etc.).
    axis : int, optional
        The axis along which to count (0 = variants, 1 = samples).

    Returns
    -------

    n : int or array
        If `axis` is None, returns the total number of Mendelian errors. If
        `axis` is specified, returns the sum along the given `axis`.

    See Also
    --------

    diploid_mendelian_error

    """

    # sum errors
    n = np.sum(diploid_mendelian_error(parental_genotypes,
                                       progeny_genotypes),
               axis=axis)

    return n


def impute_inheritance_nearest(inheritance, pos, pos_impute):
    """Impute inheritance at unknown positions, by copying from
    nearest neighbouring position where inheritance is known.

    Parameters
    ----------

    inheritance : array_like, int, shape (n_variants, n_gametes)
        An array of integers coding the allelic inheritance state at the
        known positions.
    pos : array_like, int, shape (n_variants,)
        Array of genomic positions at which `inheritance` was determined.
    pos_impute : array_like, int
        Array of positions at which to impute inheritance.

    Returns
    -------

    imputed_inheritance : ndarray, int
        An array of integers coding the imputed allelic inheritance.

    """

    # check inputs
    inheritance = np.asarray(inheritance)
    assert inheritance.ndim == 2
    pos = np.asarray(pos)
    assert pos.ndim == 1
    pos_impute = np.asarray(pos_impute)
    assert pos_impute.ndim == 1
    n_variants = pos.size
    assert inheritance.shape[0] == n_variants

    # find indices of neighbouring variants
    indices_left = np.clip(np.searchsorted(pos, pos_impute), 0, n_variants - 1)
    indices_right = np.clip(indices_left + 1, 0, n_variants - 1)
    inh_left = np.take(inheritance, indices_left, axis=0)
    inh_right = np.take(inheritance, indices_right, axis=0)

    # find positions of neighbouring variants
    pos_left = np.take(pos, indices_left)
    pos_right = np.take(pos, indices_right)

    # compute distance to neighbours
    dist_left = np.abs(pos_impute - pos_left)
    dist_right = np.abs(pos_right - pos_impute)

    # build output
    out = np.zeros_like(inh_left)
    out[dist_left < dist_right] = inh_left[dist_left < dist_right]
    out[dist_left > dist_right] = inh_right[dist_left > dist_right]

    # # use neighbour from other side where missing
    # override_left = ((dist_left < dist_right)[:, np.newaxis]
    #                  & (out == INHERIT_MISSING))
    # out[override_left] = inh_right[override_left]
    # override_right = ((dist_left > dist_right)[:, np.newaxis]
    #                   & (out == INHERIT_MISSING))
    # out[override_right] = inh_left[override_right]

    return out


def tabulate_inheritance_switches(inheritance, pos, gametes=None):
    """Tabulate switches in inheritance.

    Parameters
    ----------

    inheritance : array_like, int, shape (n_variants, n_gametes)
        An array of integers coding the allelic inheritance state at the
        known positions.
    pos : array_like, int, shape (n_variants,)
        Array of genomic positions at which `inheritance` was determined.
    gametes : sequence, length (n_gametes), optional

    Returns
    -------

    df : DataFrame
        A table of all inheritance switches observed in the gametes,
        where each row corresponds to a switch from one parental allele to
        another. The table has a hierarchical index, where the first level
        corresponds to the gamete.

    See Also
    --------

    tabulate_inheritance_blocks

    """

    # check inputs
    inheritance = np.asarray(inheritance)
    assert inheritance.ndim == 2
    n_variants, n_gametes = inheritance.shape
    pos = np.asarray(pos)
    assert pos.ndim == 1
    assert pos.size == n_variants
    if gametes is None:
        gametes = np.arange(n_gametes)
    else:
        gametes = np.asarray(gametes)
        assert gametes.ndim == 1
        assert gametes.size == n_gametes

    states = INHERIT_PARENT1, INHERIT_PARENT2
    dfs = [anhima.util.tabulate_state_transitions(inheritance[:, i],
                                                  states,
                                                  pos)
           for i in range(n_gametes)]
    df = pandas.concat(dfs, keys=gametes)
    return df


def tabulate_inheritance_blocks(inheritance, pos, gametes=None):
    """Tabulate inheritance blocks.

    Parameters
    ----------

    inheritance : array_like, int, shape (n_variants, n_gametes)
        An array of integers coding the allelic inheritance state at the
        known positions.
    pos : array_like, int, shape (n_variants,)
        Array of genomic positions at which `inheritance` was determined.
    gametes : sequence, length (n_gametes), optional

    Returns
    -------

    df : DataFrame
        A table of all inheritance blocks observed in the gametes,
        where each row corresponds to a block of inheritance from a single
        parent. The table has a hierarchical index, where the first level
        corresponds to the gamete.

    See Also
    --------

    tabulate_inheritance_switches

    """

    # check inputs
    inheritance = np.asarray(inheritance)
    assert inheritance.ndim == 2
    n_variants, n_gametes = inheritance.shape
    pos = np.asarray(pos)
    assert pos.ndim == 1
    assert pos.size == n_variants
    if gametes is None:
        gametes = np.arange(n_gametes)
    else:
        gametes = np.asarray(gametes)
        assert gametes.ndim == 1
        assert gametes.size == n_gametes

    states = INHERIT_PARENT1, INHERIT_PARENT2
    dfs = [anhima.util.tabulate_state_blocks(inheritance[:, i],
                                             states,
                                             pos)
           for i in range(n_gametes)]
    df = pandas.concat(dfs, keys=gametes)
    return df


def inheritance_block_masks(inheritance, pos):
    """Construct arrays suitable for masking the original inheritance array
    using the attributes of the inheritance blocks.

    Parameters
    ----------

    inheritance : array_like, int, shape (n_variants, n_gametes)
        An array of integers coding the allelic inheritance state at the
        known positions.
    pos : array_like, int, shape (n_variants,)
        Array of genomic positions at which `inheritance` was determined.

    Returns
    -------

    block_support : ndarray, int, shape (n_variants, n_gametes)
        An array where each item has the number of variants supporting the
        containing inheritance block.
    block_length_min : ndarray, int, shape (n_variants, n_gametes)
        An array where each item has the minimal length of the containing
        inheritance block.

    See Also
    --------

    tabulate_inheritance_blocks

    """

    # check inputs
    inheritance = np.asarray(inheritance)
    assert inheritance.ndim == 2
    n_variants, n_gametes = inheritance.shape
    pos = np.asarray(pos)
    assert pos.ndim == 1
    assert pos.size == n_variants

    df_all_blocks = tabulate_inheritance_blocks(inheritance, pos)
    block_support = np.empty(inheritance.shape, dtype=int)
    block_support.fill(-1)
    block_length_min = np.empty(inheritance.shape, dtype=pos.dtype)
    block_support.fill(-1)

    for ig in range(n_gametes):

        df_blocks = df_all_blocks.loc[ig]

        for ib, block in df_blocks.iterrows():

            block_support[block.start_max_idx:block.stop_max_idx, ig] = \
                block.support
            block_length_min[block.start_max_idx:block.stop_max_idx, ig] = \
                block.length_min

    return block_support, block_length_min



# -*- coding: utf-8 -*-
"""
Utility functions for working with genotype data.

See also the examples at:

- http://nbviewer.ipython.org/github/alimanfoo/anhima/blob/master/examples/gt.ipynb

"""  # noqa


from __future__ import division, print_function, absolute_import


# third party dependencies
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
import numexpr as ne


# internal dependencies
import anhima.loc


def is_called(genotypes):
    """Find non-missing genotype calls.

    Parameters
    ----------

    genotypes : array_like, int
        An array of shape (n_variants, n_samples, ploidy) or
        (n_variants, ploidy) or (n_samples, ploidy), where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).

    Returns
    -------

    is_called : ndarray, bool
        An array where elements are True if the genotype call is non-missing.

    See Also
    --------

    is_missing, is_hom_ref, is_het, is_hom_alt

    Notes
    -----

    Applicable to polyploid genotype calls.

    Applicable to multiallelic variants.

    """

    # check input array has 2 or more dimensions
    genotypes = np.asarray(genotypes)
    assert genotypes.ndim > 1

    # assume ploidy is fastest changing dimension
    dim_ploidy = genotypes.ndim - 1
    ploidy = genotypes.shape[dim_ploidy]
    assert ploidy > 1

    # optimisation for diploid case
    if ploidy == 2:
        allele1 = genotypes[..., 0]  # noqa
        allele2 = genotypes[..., 1]  # noqa
        ex = '(allele1 >= 0) & (allele2 >= 0)'
        out = ne.evaluate(ex)

    # general ploidy case
    else:
        out = np.all(genotypes >= 0, axis=dim_ploidy)

    return out


def count_called(genotypes, axis=None):
    """Count non-missing genotype calls.

    Parameters
    ----------

    genotypes : array_like, int
        An array of shape (n_variants, n_samples, ploidy) or
        (n_variants, ploidy) or (n_samples, ploidy), where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).
    axis : int, optional
        The axis along which to count (0 = variants, 1 = samples).

    Returns
    -------

    n : int or array
        If `axis` is None, returns the number of called (i.e., non-missing)
        genotypes. If `axis` is specified, returns the sum along the given
        `axis`.

    See Also
    --------
    is_called

    """

    # count genotypes
    n = np.sum(is_called(genotypes), axis=axis)

    return n


def is_missing(genotypes):
    """Find missing genotype calls.

    Parameters
    ----------

    genotypes : array_like, int
        An array of shape (n_variants, n_samples, ploidy) or
        (n_variants, ploidy) or (n_samples, ploidy), where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).

    Returns
    -------

    is_missing: ndarray, bool
        An array where elements are True if the genotype call is missing.

    See Also
    --------

    is_called, is_hom_ref, is_het, is_hom_alt

    Notes
    -----

    Applicable to polyploid genotype calls.

    Applicable to multiallelic variants.

    """

    # check input array has 2 or more dimensions
    genotypes = np.asarray(genotypes)
    assert genotypes.ndim > 1

    # assume ploidy is fastest changing dimension
    dim_ploidy = genotypes.ndim - 1
    ploidy = genotypes.shape[dim_ploidy]
    assert ploidy > 1

    # optimisation for diploid case
    if ploidy == 2:
        allele1 = genotypes[..., 0]  # noqa
        allele2 = genotypes[..., 1]  # noqa
        ex = '(allele1 < 0) | (allele2 < 0)'
        out = ne.evaluate(ex)

    # general ploidy case
    else:
        out = np.any(genotypes < 0, axis=dim_ploidy)

    return out


def count_missing(genotypes, axis=None):
    """Count non-missing genotype calls.

    Parameters
    ----------

    genotypes : array_like, int
        An array of shape (n_variants, n_samples, ploidy) or
        (n_variants, ploidy) or (n_samples, ploidy), where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).
    axis : int, optional
        The axis along which to count (0 = variants, 1 = samples).

    Returns
    -------

    n : int or array
        If `axis` is None, returns the number of missing genotypes. If `axis`
        is specified, returns the sum along the given `axis`.

    See Also
    --------
    is_missing

    """

    # count genotypes
    n = np.sum(is_missing(genotypes), axis=axis)

    return n


def is_hom(genotypes):
    """Find homozygous genotype calls.

    Parameters
    ----------

    genotypes : array_like, int
        An array of shape (n_variants, n_samples, ploidy) or
        (n_variants, ploidy) or (n_samples, ploidy), where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).

    Returns
    -------

    is_hom : ndarray, bool
        An array where elements are True if the genotype call is homozygous.

    See Also
    --------

    is_called, is_missing, is_hom_ref, is_hom_alt

    Notes
    -----

    Applicable to polyploid genotype calls.

    Applicable to multiallelic variants.

    """

    # check input array has 2 or more dimensions
    genotypes = np.asarray(genotypes)
    assert genotypes.ndim > 1

    # assume ploidy is fastest changing dimension
    dim_ploidy = genotypes.ndim - 1
    ploidy = genotypes.shape[dim_ploidy]
    assert ploidy > 1

    # optimisation for diploid case
    if ploidy == 2:
        allele1 = genotypes[..., 0]  # noqa
        allele2 = genotypes[..., 1]  # noqa
        ex = '(allele1 >= 0) & (allele1 == allele2)'
        out = ne.evaluate(ex)

    # general ploidy case
    else:
        allele1 = genotypes[..., 0, np.newaxis]  # noqa
        other_alleles = genotypes[..., 1:]  # noqa
        ex = '(allele1 >= 0) & (allele1 == other_alleles)'
        out = np.all(ne.evaluate(ex), axis=dim_ploidy)

    return out


def count_hom(genotypes, axis=None):
    """Count homozygous genotype calls.

    Parameters
    ----------

    genotypes : array_like, int
        An array of shape (n_variants, n_samples, ploidy) or
        (n_variants, ploidy) or (n_samples, ploidy), where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).
    axis : int, optional
        The axis along which to count (0 = variants, 1 = samples).

    Returns
    -------

    n : int or array
        If `axis` is None, returns the number of homozygous genotypes. If
        `axis` is specified, returns the sum along the given `axis`.

    See Also
    --------
    is_hom

    """

    # count genotypes
    n = np.sum(is_hom(genotypes), axis=axis)

    return n


def is_het(genotypes):
    """Find heterozygous genotype calls.

    Parameters
    ----------

    genotypes : array_like, int
        An array of shape (n_variants, n_samples, ploidy) or
        (n_variants, ploidy) or (n_samples, ploidy), where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).

    Returns
    -------

    is_het : ndarray, bool
        An array where elements are True if the genotype call is heterozygous.

    See Also
    --------

    is_called, is_missing, is_hom_ref, is_hom_alt

    Notes
    -----

    Applicable to polyploid genotype calls, although note that all
    types of heterozygous genotype (i.e., anything not completely
    homozygous) will give an element value of True.

    Applicable to multiallelic variants, although note that the element value
    will be True in any case where the two alleles in a genotype are
    different, e.g., (0, 1), (0, 2), (1, 2), etc.

    """

    # check input array has 2 or more dimensions
    genotypes = np.asarray(genotypes)
    assert genotypes.ndim > 1

    # assume ploidy is fastest changing dimension
    dim_ploidy = genotypes.ndim - 1
    ploidy = genotypes.shape[dim_ploidy]
    assert ploidy > 1

    # optimisation for diploid case
    if ploidy == 2:
        allele1 = genotypes[..., 0]
        allele2 = genotypes[..., 1]
        out = allele1 != allele2

    # general ploidy case
    else:
        allele1 = genotypes[..., 0, np.newaxis]
        other_alleles = genotypes[..., 1:]
        out = np.any(allele1 != other_alleles, axis=dim_ploidy)

    return out


def count_het(genotypes, axis=None):
    """Count heterozygous genotype calls.

    Parameters
    ----------

    genotypes : array_like, int
        An array of shape (n_variants, n_samples, ploidy) or
        (n_variants, ploidy) or (n_samples, ploidy), where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).
    axis : int, optional
        The axis along which to count (0 = variants, 1 = samples).

    Returns
    -------

    n : int or array
        If `axis` is None, returns the number of heterozygous genotypes. If
        `axis` is specified, returns the sum along the given `axis`.

    See Also
    --------
    is_het

    """

    # count genotypes
    n = np.sum(is_het(genotypes), axis=axis)

    return n


def is_hom_ref(genotypes):
    """Find homozygous reference genotype calls.

    Parameters
    ----------

    genotypes : array_like, int
        An array of shape (n_variants, n_samples, ploidy) or
        (n_variants, ploidy) or (n_samples, ploidy), where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).

    Returns
    -------

    is_hom_ref : ndarray, bool
        An array where elements are True if the genotype call is homozygous
        reference.

    See Also
    --------
    is_called, is_missing, is_het, is_hom_alt

    Notes
    -----

    Applicable to polyploid genotype calls.

    Applicable to multiallelic variants.

    """

    # check input array has 2 or more dimensions
    genotypes = np.asarray(genotypes)
    assert genotypes.ndim > 1

    # assume ploidy is fastest changing dimension
    dim_ploidy = genotypes.ndim - 1
    ploidy = genotypes.shape[dim_ploidy]
    assert ploidy > 1

    # optimisation for diploid case
    if ploidy == 2:
        allele1 = genotypes[..., 0]  # noqa
        allele2 = genotypes[..., 1]  # noqa
        ex = '(allele1 == 0) & (allele2 == 0)'
        out = ne.evaluate(ex)

    # general ploidy case
    else:
        out = np.all(genotypes == 0, axis=dim_ploidy)

    return out


def count_hom_ref(genotypes, axis=None):
    """Count homozygous reference genotype calls.

    Parameters
    ----------

    genotypes : array_like, int
        An array of shape (n_variants, n_samples, ploidy) or
        (n_variants, ploidy) or (n_samples, ploidy), where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).
    axis : int, optional
        The axis along which to count (0 = variants, 1 = samples).

    Returns
    -------

    n : int or array
        If `axis` is None, returns the number of homozygous
        reference genotypes. If `axis` is specified, returns the sum along
        the given `axis`.

    See Also
    --------
    is_hom_ref

    """

    # count genotypes
    n = np.sum(is_hom_ref(genotypes), axis=axis)

    return n


def is_hom_alt(genotypes):
    """Find homozygous non-reference genotype calls.

    Parameters
    ----------

    genotypes : array_like, int
        An array of shape (n_variants, n_samples, ploidy) or
        (n_variants, ploidy) or (n_samples, ploidy), where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).

    Returns
    -------

    is_hom_alt : ndarray, bool
        An array where elements are True if the genotype call is homozygous
        non-reference.

    See Also
    --------

    is_called, is_missing, is_hom_ref, is_het

    Notes
    -----

    Applicable to polyploid genotype calls.

    Applicable to multiallelic variants.

    """

    # check input array has 2 or more dimensions
    genotypes = np.asarray(genotypes)
    assert genotypes.ndim > 1

    # assume ploidy is fastest changing dimension
    dim_ploidy = genotypes.ndim - 1
    ploidy = genotypes.shape[dim_ploidy]
    assert ploidy > 1

    # optimisation for diploid case
    if ploidy == 2:
        allele1 = genotypes[..., 0]  # noqa
        allele2 = genotypes[..., 1]  # noqa
        ex = '(allele1 > 0) & (allele1 == allele2)'
        out = ne.evaluate(ex)

    # general ploidy case
    else:
        allele1 = genotypes[..., 0, np.newaxis]  # noqa
        other_alleles = genotypes[..., 1:]  # noqa
        ex = '(allele1 > 0) & (allele1 == other_alleles)'
        out = np.all(ne.evaluate(ex), axis=dim_ploidy)

    return out


def count_hom_alt(genotypes, axis=None):
    """Count homozygous non-reference genotype calls.

    Parameters
    ----------

    genotypes : array_like, int
        An array of shape (n_variants, n_samples, ploidy) or
        (n_variants, ploidy) or (n_samples, ploidy), where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).
    axis : int, optional
        The axis along which to count (0 = variants, 1 = samples).

    Returns
    -------

    n : int or array
        If `axis` is None, returns the number of homozygous non-reference
        genotypes. If `axis` is specified, returns the sum along the given
        `axis`.

    See Also
    --------
    is_hom_alt

    """

    # count genotypes
    n = np.sum(is_hom_alt(genotypes), axis=axis)

    return n


def max_allele(genotypes, axis=None):
    """
    Return the highest allele index.

    Parameters
    ----------

    genotypes : array_like
        An array of shape (n_variants, n_samples, ploidy) where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).
    axis : int, optional
        The axis along which to determine the maximum (0 = variants,
        1 = samples). If not given, return the highest overall.

    Returns
    -------

    n : int
        The value of the highest allele index present in the genotypes array.

    """

    return np.amax(genotypes, axis=axis)


def as_haplotypes(genotypes):
    """Reshape an array of genotypes to view it as haplotypes by dropping the
    ploidy dimension.

    Parameters
    ----------

    genotypes : array_like, int
        An array of shape (n_variants, n_samples, ploidy) where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).

    Returns
    -------

    haplotypes : ndarray
        An array of shape (n_variants, n_samples * ploidy).

    Notes
    -----

    Note that if genotype calls are unphased, the haplotypes returned by this
    function will bear no resemblance to the true haplotypes.

    Applicable to polyploid genotype calls.

    Applicable to multiallelic variants.

    """

    # check input array
    genotypes = np.asarray(genotypes)
    assert genotypes.ndim == 3

    # reshape, preserving size of first dimension
    newshape = (genotypes.shape[0], -1)
    haplotypes = np.reshape(genotypes, newshape)

    return haplotypes


def as_n_alt(genotypes):
    """Transform genotypes as the number of non-reference alleles.

    Parameters
    ----------

    genotypes : array_like, int
        An array of shape (n_variants, n_samples, ploidy) or
        (n_variants, ploidy) or (n_samples, ploidy), where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).

    Returns
    -------

    gn : ndarray, uint8
        An array where each genotype is coded as a single integer counting
        the number of alternate alleles.

    See Also
    --------

    as_012

    Notes
    -----

    Applicable to polyploid genotype calls.

    Applicable to multiallelic variants, although this function simply
    counts the number of non-reference alleles, it makes no distinction
    between different non-reference alleles.

    Note that this function returns 0 for missing genotype calls **and** for
    homozygous reference genotype calls, because in both cases the number of
    non-reference alleles is zero.

    """

    # check input array
    genotypes = np.asarray(genotypes)
    assert genotypes.ndim > 1

    # assume ploidy is fastest changing dimension
    dim_ploidy = genotypes.ndim - 1

    # count number of alternate alleles
    gn = np.empty(genotypes.shape[:-1], dtype='u1')
    np.sum(genotypes > 0, axis=dim_ploidy, out=gn)

    return gn


def as_012(genotypes, fill=-1):
    """Transform genotypes recoding homozygous reference calls a
    0, heterozygous calls as 1, homozygous non-reference calls as 2, and
    missing calls as -1.

    Parameters
    ----------

    genotypes : array_like, int
        An array of shape (n_variants, n_samples, ploidy) or
        (n_variants, ploidy) or (n_samples, ploidy), where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).
    fill : int, optional
        Default value for missing calls.

    Returns
    -------

    gn : ndarray, int8
        An array where each genotype is coded as a single integer as
        described above.

    See Also
    --------

    as_nalt

    Notes
    -----

    Applicable to polyploid genotype calls, although note that all
    types of heterozygous genotype (i.e., anything not completely
    homozygous) will be coded as 1.

    Applicable to multiallelic variants, although note the following.
    All heterozygous genotypes, e.g., (0, 1), (0, 2), (1, 2), ..., will be
    coded as 1. All homozygous non-reference genotypes, e.g., (1, 1), (2, 2),
    ..., will be coded as 2.

    """

    # check input array
    genotypes = np.asarray(genotypes)
    assert genotypes.ndim > 1

    # set up output array
    gn = np.empty(genotypes.shape[:-1], dtype='i1')
    gn.fill(fill)

    # determine genotypes
    gn[is_hom_ref(genotypes)] = 0
    gn[is_het(genotypes)] = 1
    gn[is_hom_alt(genotypes)] = 2

    return gn


def as_allele_counts(genotypes, alleles=None):
    """Transform genotypes into allele counts per call.

    Parameters
    ----------

    genotypes : array_like, int
        An array of shape (n_variants, n_samples, ploidy) or
        (n_variants, ploidy) or (n_samples, ploidy), where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).
    alleles : sequence of ints, optional
        The alleles to count. If not specified, all alleles will be counted.

    Returns
    -------

    gac : ndarray, uint8
        An array where the ploidy dimension has been replaced by counts of
        each allele.

    """

    # check inputs
    genotypes = np.asarray(genotypes)
    assert genotypes.ndim > 1
    dim_ploidy = genotypes.ndim - 1
    if alleles is None:
        m = np.amax(genotypes)
        alleles = range(m+1)

    # count alleles along ploidy dimension
    gac = np.zeros(genotypes.shape[:-1] + (len(alleles),), dtype='u1')
    for i, allele in enumerate(alleles):
        np.sum(genotypes == allele, axis=dim_ploidy, out=gac[..., i])

    return gac


def pack_diploid(genotypes):
    """
    Pack diploid genotypes into a single byte for each genotype,
    using the left-most 4 bits for the first allele and the right-most 4 bits
    for the second allele. Allows single byte encoding of diploid genotypes
    for variants with up to 15 alleles.

    Parameters
    ----------

    genotypes : array_like, int
        An array of shape (n_variants, n_samples, ploidy) or
        (n_variants, ploidy) or (n_samples, ploidy), where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).

    Returns
    -------

    packed : ndarray, int8
        An array of genotypes where the ploidy dimension has been collapsed
        by bit packing the two alleles for each genotype into a single byte.

    See Also
    --------

    unpack_diploid_genotypes

    """

    # check inputs
    genotypes = np.asarray(genotypes)
    assert genotypes.ndim > 1
    assert np.amax(genotypes) < 15, 'max allele is 14'
    assert np.amin(genotypes) > -2, 'min allele is -1'
    genotypes = genotypes.astype('u1')

    # add 1 to handle missing alleles coded as -1
    genotypes = genotypes + 1

    # left shift first allele by 4 bits
    a1 = np.left_shift(genotypes[..., 0], 4)

    # mask left-most 4 bits to ensure second allele doesn't clash with first
    # allele
    a2 = np.bitwise_and(genotypes[..., 1], 15)

    # pack them
    packed = np.bitwise_or(a1, a2)

    # rotate round so that hom ref calls are encoded as 0, better for sparse
    # matrices
    packed -= 17

    return packed


def unpack_diploid(packed):
    """
    Unpack an array of diploid genotypes that have been bit packed into
    single bytes.

    Parameters
    ----------

    packed : array_like
        An array of genotypes where the ploidy dimension has been collapsed
        by bit packing the two alleles for each genotype into a single byte.

    Returns
    -------

    genotypes : ndarray, int8
        An array of genotypes where the ploidy dimension has been restored by
        unpacking the input array.

    See Also
    --------

    pack_diploid_genotypes

    """

    # check inputs
    packed = np.asarray(packed).astype('u1')
    assert 1 <= packed.ndim <= 2

    # rotate back round so missing calls are encoded as 0
    packed = packed + 17

    # set up output array
    genotypes = np.empty(packed.shape + (2,), dtype='u1')
    a1 = genotypes[..., 0]
    a2 = genotypes[..., 1]

    # right shift 4 bits to extract first allele
    np.right_shift(packed, 4, out=a1)

    # mask left-most 4 bits to extract second allele
    np.bitwise_and(packed, 15, out=a2)

    # subtract 1 to restore coding of missing alleles as -1
    genotypes -= 1
    genotypes = genotypes.astype('i1')

    return genotypes


# packed representation of some common diploid genotypes
BHOM00 = 0
BHET01 = 1
BHOM11 = 17
BMISSING = 239


def count_genotypes(gn, t, axis=None):
    """Count genotypes of a given type.

    Parameters
    ----------

    gn : array_like, int
        An array of shape (n_variants, n_samples) or (n_variants,) or
        (n_samples,) where each element is a genotype called coded as a
        single integer.
    t : int
        The genotype to count.
    axis : int, optional
        The axis along which to count (0 = variants, 1 = samples).

    Returns
    -------

    n : int or array
        If `axis` is None, returns the total number of matching genotypes. If
        `axis` is specified, returns the sum along the given `axis`.

    """

    # normalise inputs
    gn = np.asarray(gn)

    # count genotypes
    n = np.sum(gn == t, axis=axis)

    return n


def windowed_genotype_counts(pos, gn, t, window_size, start_position=None,
                             stop_position=None):
    """Count genotype calls of a given type for a single sample in
    non-overlapping windows over the genome.

    Parameters
    ----------

    pos : array_like, int
        A sorted 1-dimensional array of genomic positions from a single
        chromosome/contig.
    gn : array_like, int
        A 1-D array of genotypes for a single sample, where each genotype is
        coded as a single integer.
    t : int
        The genotype to count.
    window_size : int
        The size in base-pairs of the windows.
    start_position : int, optional
        The start position for the region over which to work.
    stop_position : int, optional
        The stop position for the region over which to work.

    Returns
    -------

    counts : ndarray, int
        Genotype counts for each window.
    bin_edges : ndarray, float
        The edges of the windows.

    See Also
    --------

    as_diploid_012, as_n_alt, windowed_genotype_density, windowed_genotype_rate

    """

    # check input array
    gn = np.asarray(gn)
    assert gn.ndim == 1

    # find matching genotypes
    values = gn == t

    # computed binned statistic
    counts, bin_edges = anhima.loc.windowed_statistic(
        pos, values=values, statistic='sum', window_size=window_size,
        start_position=start_position, stop_position=stop_position
    )

    return counts, bin_edges


def windowed_genotype_density(pos, gn, t, window_size, start_position=None,
                              stop_position=None):
    """Compute per-base-pair density of genotype calls of a given type for a
    single sample in non-overlapping windows over the genome.

    Parameters
    ----------

    pos : array_like, int
        A sorted 1-dimensional array of genomic positions from a single
        chromosome/contig.
    gn : array_like, int
        A 1-D array of genotypes for a single sample, where each genotype is
        coded as a single integer.
    t : int
        The genotype to count.
    window_size : int
        The size in base-pairs of the windows.
    start_position : int, optional
        The start position for the region over which to work.
    stop_position : int, optional
        The stop position for the region over which to work.

    Returns
    -------

    density : ndarray, float
        Genotype density for each window.
    bin_edges : ndarray, float
        The edges of the windows.

    See Also
    --------

    as_diploid_012, as_n_alt, windowed_genotype_counts, windowed_genotype_rate

    """

    counts, bin_edges = windowed_genotype_counts(pos, gn, t,
                                                 window_size=window_size,
                                                 start_position=start_position,
                                                 stop_position=stop_position)
    density = counts / np.diff(bin_edges)
    return density, bin_edges


def windowed_genotype_rate(pos, gn, t, window_size, start_position=None,
                           stop_position=None):
    """Compute per-variant rate of genotype calls of a given type for a
    single sample in non-overlapping windows over the genome.

    Parameters
    ----------

    pos : array_like, int
        A sorted 1-dimensional array of genomic positions from a single
        chromosome/contig.
    gn : array_like, int
        A 1-D array of genotypes for a single sample, where each genotype is
        coded as a single integer.
    t : int
        The genotype to count.
    window_size : int
        The size in base-pairs of the windows.
    start_position : int, optional
        The start position for the region over which to work.
    stop_position : int, optional
        The stop position for the region over which to work.

    Returns
    -------

    rate : ndarray, float
        Per-variant rate for each window.
    bin_edges : ndarray, float
        The edges of the windows.

    See Also
    --------

    as_diploid_012, as_n_alt, windowed_genotype_counts,
    windowed_genotype_density

    """

    variant_counts, _ = anhima.loc.windowed_variant_counts(
        pos, window_size, start_position=start_position,
        stop_position=stop_position
    )
    counts, bin_edges = windowed_genotype_counts(
        pos, gn, t, window_size=window_size, start_position=start_position,
        stop_position=stop_position
    )
    rate = counts / variant_counts
    return rate, bin_edges


def plot_windowed_genotype_counts(pos, gn, t, window_size, start_position=None,
                                  stop_position=None, ax=None,
                                  plot_kwargs=None):
    """Plots counts of genotype calls of a given type for a single sample in
    non-overlapping windows over the genome.

    Parameters
    ----------

    pos : array_like, int
        A sorted 1-dimensional array of genomic positions from a single
        chromosome/contig.
    gn : array_like, int
        A 1-D array of genotypes for a single sample, where each genotype is
        coded as a single integer.
    t : int
        The genotype to count.
    window_size : int
        The size in base-pairs of the windows.
    start_position : int, optional
        The start position for the region over which to work.
    stop_position : int, optional
        The stop position for the region over which to work.
    ax : axes, optional
        The axes on which to draw. If not provided, a new figure will be
        created.
    plot_kwargs : dict-like
        Additional keyword arguments passed through to `plt.plot`.

    Returns
    -------

    ax : axes
        The axes on which the plot was drawn.

    """

    # set up axes
    if ax is None:
        x = plt.rcParams['figure.figsize'][0]
        fig = plt.figure(figsize=(x, x//3))
        ax = fig.add_subplot(111)

    # count genotypes
    y, bin_edges = windowed_genotype_counts(pos, gn, t, window_size,
                                            start_position=start_position,
                                            stop_position=stop_position)
    x = (bin_edges[1:] + bin_edges[:-1])/2

    # plot data
    if plot_kwargs is None:
        plot_kwargs = dict()
    plot_kwargs.setdefault('linestyle', '-')
    plot_kwargs.setdefault('marker', None)
    ax.plot(x, y, label=t, **plot_kwargs)

    # tidy up
    ax.set_ylim(bottom=0)
    ax.set_xlabel('position')
    ax.set_ylabel('counts')
    if start_position is None:
        start_position = np.min(pos)
    if stop_position is None:
        stop_position = np.max(pos)
    ax.set_xlim(start_position, stop_position)

    return ax


def plot_windowed_genotype_density(pos, gn, t, window_size,
                                   start_position=None,
                                   stop_position=None, ax=None,
                                   plot_kwargs=None):
    """Plots per-base-pair density of genotype calls of a given type for a
    single sample in non-overlapping windows over the genome.

    Parameters
    ----------

    pos : array_like, int
        A sorted 1-dimensional array of genomic positions from a single
        chromosome/contig.
    gn : array_like, int
        A 1-D array of genotypes for a single sample, where each genotype is
        coded as a single integer.
    t : int
        The genotype to count.
    window_size : int
        The size in base-pairs of the windows.
    start_position : int, optional
        The start position for the region over which to work.
    stop_position : int, optional
        The stop position for the region over which to work.
    ax : axes, optional
        The axes on which to draw. If not provided, a new figure will be
        created.
    plot_kwargs : dict-like
        Additional keyword arguments passed through to `plt.plot`.

    Returns
    -------

    ax : axes
        The axes on which the plot was drawn.

    """

    # set up axes
    if ax is None:
        x = plt.rcParams['figure.figsize'][0]
        fig = plt.figure(figsize=(x, x//3))
        ax = fig.add_subplot(111)

    # count genotypes
    y, bin_edges = windowed_genotype_density(pos, gn, t, window_size,
                                             start_position=start_position,
                                             stop_position=stop_position)
    x = (bin_edges[1:] + bin_edges[:-1])/2

    # plot data
    if plot_kwargs is None:
        plot_kwargs = dict()
    plot_kwargs.setdefault('linestyle', '-')
    plot_kwargs.setdefault('marker', None)
    ax.plot(x, y, label=t, **plot_kwargs)

    # tidy up
    ax.set_ylim(bottom=0)
    ax.set_xlabel('position')
    ax.set_ylabel('density')
    if start_position is None:
        start_position = np.min(pos)
    if stop_position is None:
        stop_position = np.max(pos)
    ax.set_xlim(start_position, stop_position)

    return ax


def plot_windowed_genotype_rate(pos, gn, t, window_size,
                                start_position=None,
                                stop_position=None, ax=None,
                                plot_kwargs=None):
    """Plots per-variant rate of genotype calls of a given type for a
    single sample in non-overlapping windows over the genome.

    Parameters
    ----------

    pos : array_like, int
        A sorted 1-dimensional array of genomic positions from a single
        chromosome/contig.
    gn : array_like, int
        A 1-D array of genotypes for a single sample, where each genotype is
        coded as a single integer.
    t : int
        The genotype to count.
    window_size : int
        The size in base-pairs of the windows.
    start_position : int, optional
        The start position for the region over which to work.
    stop_position : int, optional
        The stop position for the region over which to work.
    ax : axes, optional
        The axes on which to draw. If not provided, a new figure will be
        created.
    plot_kwargs : dict-like
        Additional keyword arguments passed through to `plt.plot`.

    Returns
    -------

    ax : axes
        The axes on which the plot was drawn.

    """

    # set up axes
    if ax is None:
        x = plt.rcParams['figure.figsize'][0]
        fig = plt.figure(figsize=(x, x//3))
        ax = fig.add_subplot(111)

    # count genotypes
    y, bin_edges = windowed_genotype_rate(pos, gn, t, window_size,
                                          start_position=start_position,
                                          stop_position=stop_position)
    x = (bin_edges[1:] + bin_edges[:-1])/2

    # plot data
    if plot_kwargs is None:
        plot_kwargs = dict()
    plot_kwargs.setdefault('linestyle', '-')
    plot_kwargs.setdefault('marker', None)
    ax.plot(x, y, label=t, **plot_kwargs)

    # tidy up
    ax.set_ylim(bottom=0)
    ax.set_xlabel('position')
    ax.set_ylabel('per variant rate')
    if start_position is None:
        start_position = np.min(pos)
    if stop_position is None:
        stop_position = np.max(pos)
    ax.set_xlim(start_position, stop_position)

    return ax


def plot_discrete_calldata(a, labels=None, colors='wbgrcmyk', states=None,
                           ax=None, pcolormesh_kwargs=None):
    """
    Plot a color grid from discrete calldata (e.g., genotypes).

    Parameters
    ----------

    a : array_like, int, shape (n_variants, n_samples)
        2-dimensional array of integers containing the call data to plot.
    labels : sequence of strings, optional
        Axis labels (e.g., sample IDs).
    colors : sequence, optional
        Colors to use for different values of the array.
    states : sequence, optional
        Manually specify discrete calldata states (if not given will be
        determined from the data).
    ax : axes, optional
        The axes on which to draw. If not provided, a new figure will be
        created.
    pcolormesh_kwargs : dict-like, optional
        Additional keyword arguments passed through to `plt.pcolormesh`.

    Returns
    -------

    ax : axes
        The axes on which the plot was drawn.

    """

    # check input array
    a = np.asarray(a)
    assert a.ndim == 2

    # set up axes
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)

    # determine discrete states
    if states is None:
        states = np.unique(a)

    # determine colors for states
    colors = colors[:np.max(states)-np.min(states)+1]

    # plotting defaults
    if pcolormesh_kwargs is None:
        pcolormesh_kwargs = dict()
    pcolormesh_kwargs.setdefault('cmap', mpl.colors.ListedColormap(colors))
    pcolormesh_kwargs.setdefault(
        'norm', plt.Normalize(np.min(states), np.max(states)+1)
    )

    # plot the colormesh
    ax.pcolormesh(a.T, **pcolormesh_kwargs)

    # tidy up
    ax.set_xlim(0, a.shape[0])
    ax.set_ylim(0, a.shape[1])
    ax.set_xticks([])
    if labels is not None:
        ax.set_yticks(np.arange(a.shape[1]) + .5)
        ax.set_yticklabels(labels, rotation=0)
    else:
        ax.set_yticks([])

    return ax


def plot_continuous_calldata(a, labels=None, ax=None, pcolormesh_kwargs=None):
    """
    Plot a color grid from continuous calldata (e.g., DP).

    Parameters
    ----------

    a : array_like, shape (n_variants, n_samples)
        2-dimensional array of integers or floats containing the call data to
        plot.
    labels : sequence of strings, optional
        Axis labels (e.g., sample IDs).
    ax : axes, optional
        The axes on which to draw. If not provided, a new figure will be
        created.
    pcolormesh_kwargs : dict-like, optional
        Additional keyword arguments passed through to `plt.pcolormesh`.

    Returns
    -------

    ax : axes
        The axes on which the plot was drawn.

    """

    # check input array
    a = np.asarray(a)
    assert a.ndim == 2

    # set up axes
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)

    # plotting defaults
    if pcolormesh_kwargs is None:
        pcolormesh_kwargs = dict()
    pcolormesh_kwargs.setdefault('cmap', 'jet')

    # plot the color mesh
    ax.pcolormesh(a.T, **pcolormesh_kwargs)

    # tidy up
    ax.set_xlim(0, a.shape[0])
    ax.set_ylim(0, a.shape[1])
    ax.set_xticks([])
    if labels is not None:
        ax.set_yticks(np.arange(a.shape[1]) + .5)
        ax.set_yticklabels(labels, rotation=0)
    else:
        ax.set_yticks([])

    return ax


def plot_diploid_genotypes(gn,
                           labels=None,
                           colors='wbgr',
                           states=(-1, 0, 1, 2),
                           ax=None,
                           colormesh_kwargs=None):
    """Plot diploid genotypes as a color grid.

    Parameters
    ----------

    gn : array_like, int, shape (n_variants, n_samples)
        An array where each genotype is coded as a single integer as
        described above.
    labels : sequence of strings, optional
        Axis labels (e.g., sample IDs).
    colors : sequence, optional
        Colors to use for different values of the array.
    states : sequence, optional
        Manually specify discrete calldata states (if not given will be
        determined from the data).
    ax : axes, optional
        The axes on which to draw. If not provided, a new figure will be
        created.
    colormesh_kwargs : dict-like
        Additional keyword arguments passed through to `plt.pcolormesh`.

    Returns
    -------

    ax : axes
        The axes on which the plot was drawn.

    """

    return plot_discrete_calldata(gn, labels=labels, colors=colors,
                                  states=states, ax=ax,
                                  pcolormesh_kwargs=colormesh_kwargs)


def plot_genotype_counts_by_sample(gn, states=(-1, 0, 1, 2),
                                   colors='wbgr', labels=None,
                                   ax=None, width=1, orientation='vertical',
                                   bar_kwargs=None):
    """Plot a bar graph of genotype counts by sample.

    Parameters
    ----------

    gn : array_like, int, shape (n_variants, n_samples)
        An array where each genotype is coded as a single integer as
        described above.
    states : sequence, optional
        The genotype states to count.
    colors : sequence, optional
        Colors to use for corresponding states.
    labels : sequence of strings, optional
        Axis labels (e.g., sample IDs).
    ax : axes, optional
        The axes on which to draw. If not provided, a new figure will be
        created.
    width : float, optional
        Width of the bars (will be used as height if `orientation` ==
        'horizontal').
    orientation : {'horizontal', 'vertical'}
        Which type of bar to plot.
    bar_kwargs : dict-like
        Additional keyword arguments passed through to `plt.bar`.

    Returns
    -------

    ax : axes
        The axes on which the plot was drawn.

    """

    # check input array
    gn = np.asarray(gn)
    assert gn.ndim == 2
    n_variants = gn.shape[0]
    n_samples = gn.shape[1]

    # check orientation
    assert orientation in ('vertical', 'horizontal')

    # set up axes
    if ax is None:
        fig, ax = plt.subplots()

    # determine bar positions
    x = np.arange(n_samples)

    # plot bars for each type
    if bar_kwargs is None:
        bar_kwargs = dict()
    bar_kwargs.setdefault('linewidth', 0)
    yc = None
    for t, c in zip(states, colors):

        # count genotypes
        y = count_genotypes(gn, t, axis=0)

        # plot as bar
        if orientation == 'vertical':
            ax.bar(x, y, width=width, bottom=yc, color=c, label=t,
                   **bar_kwargs)
        else:
            ax.barh(x, y, height=width, left=yc, color=c, label=t,
                    **bar_kwargs)

        # keep cumulative count
        if yc is None:
            yc = y
        else:
            yc += y

    # tidy up
    # TODO code smells

    # set plot limits
    if orientation == 'vertical':
        ax.set_ylim(0, n_variants)
        ax.set_xlim(0, n_samples)
    else:
        ax.set_xlim(0, n_variants)
        ax.set_ylim(0, n_samples)

    # determine tick labels
    if labels:
        if orientation == 'vertical':
            ax.set_xticks(range(n_samples))
            ax.set_xticklabels(labels)
        else:
            ax.set_yticks(range(n_samples))
            ax.set_yticklabels(labels)
    else:
        if orientation == 'vertical':
            ax.set_xticks([])
        else:
            ax.set_yticks([])

    return ax


def plot_genotype_counts_by_variant(gn, states=(-1, 0, 1, 2),
                                    colors='wbgr', ax=None, width=1,
                                    orientation='vertical',
                                    bar_kwargs=None):
    """Plot a bar graph of genotype counts by variant.

    Parameters
    ----------

    gn : array_like, int, shape (n_variants, n_samples)
        An array where each genotype is coded as a single integer as
        described above.
    states : sequence, optional
        The genotype states to count.
    colors : sequence, optional
        Colors to use for corresponding states.
    ax : axes, optional
        The axes on which to draw. If not provided, a new figure will be
        created.
    width : float, optional
        Width of the bars (will be used as height if `orientation` ==
        'horizontal').
    orientation : {'horizontal', 'vertical'}
        Which type of bar to plot.
    bar_kwargs : dict-like
        Additional keyword arguments passed through to `plt.bar`.

    Returns
    -------

    ax : axes
        The axes on which the plot was drawn.

    """

    # check input array
    gn = np.asarray(gn)
    assert gn.ndim == 2
    n_variants = gn.shape[0]
    n_samples = gn.shape[1]

    # check orientation
    assert orientation in ('vertical', 'horizontal')

    # set up axes
    if ax is None:
        fig, ax = plt.subplots()

    # determine bar positions
    x = np.arange(n_variants)

    # plot bars for each type
    if bar_kwargs is None:
        bar_kwargs = dict()
    bar_kwargs.setdefault('linewidth', 0)
    yc = None
    for t, c in zip(states, colors):

        # count genotypes
        y = count_genotypes(gn, t, axis=1)

        # plot as bar
        if orientation == 'vertical':
            ax.bar(x, y, width=width, bottom=yc, color=c, label=t,
                   **bar_kwargs)
        else:
            ax.barh(x, y, height=width, left=yc, color=c, label=t,
                    **bar_kwargs)

        # keep cumulative count
        if yc is None:
            yc = y
        else:
            yc += y

    # tidy up
    if orientation == 'vertical':
        ax.set_xticks([])
        ax.set_xlim(0, n_variants)
        ax.set_ylim(0, n_samples)
    else:
        ax.set_yticks([])
        ax.set_ylim(0, n_variants)
        ax.set_xlim(0, n_samples)

    return ax


def plot_continuous_calldata_by_sample(a, labels=None,
                                       ax=None,
                                       orientation='vertical',
                                       boxplot_kwargs=None):
    """Plot a boxplot of continuous call data (e.g., DP) by sample.

    Parameters
    ----------

    a : array_like, shape (n_variants, n_samples)
        2-dimensional array of integers or floats containing the call data to
        plot.
    labels : sequence of strings, optional
        Axis labels (e.g., sample IDs).
    ax : axes, optional
        The axes on which to draw. If not provided, a new figure will be
        created.
    orientation : {'horizontal', 'vertical'}
        Which type of bar to plot.
    boxplot_kwargs : dict-like
        Additional keyword arguments passed through to `plt.boxplot`.

    Returns
    -------

    ax : axes
        The axes on which the plot was drawn.

    """

    # check input array
    a = np.asarray(a)
    assert a.ndim == 2

    # check orientation
    assert orientation in ('vertical', 'horizontal')
    vert = orientation == 'vertical'

    # set up axes
    if ax is None:
        fig, ax = plt.subplots()

    # plot
    if boxplot_kwargs is None:
        boxplot_kwargs = dict()
    ax.boxplot(a, vert=vert, **boxplot_kwargs)

    # tidy up
    n_samples = a.shape[1]
    if labels:
        if orientation == 'vertical':
            ax.set_xticks(range(n_samples))
            ax.set_xticklabels(labels)
        else:
            ax.set_yticks(range(n_samples))
            ax.set_yticklabels(labels)
    else:
        if orientation == 'vertical':
            ax.set_xticks([])
        else:
            ax.set_yticks([])

    return ax
